#%%
"""
Created on Thu Nov 30 2018
Merton Model and convergence obtained with the COS method
@author: Lech A. Grzelak
"""
import numpy as np
import matplotlib.pyplot as plt
from scipy.stats import norm
import enum
import scipy.optimize as optimize
from numba import jit


def call_put_option_price_cos_method(cf, cp, S0, r, tau, K, N, L):
    # cf - Characteristic function as a functon, in the book denoted by phi
    # cp - C for call and P for put
    # S0 - Initial stock price
    # r - Interest rate (constant)
    # tau - Time to maturity
    # K - List of strikes
    # N - Number of expansion terms
    # L - Size of truncation domain (typ.:L=8 or L=10)

    # Reshape K to become a column vector
    if K is not np.array:
        K = np.array(K).reshape([len(K),1])

    x0 = np.log(S0 / np.array(K)).reshape(-1, 1)

    # Truncation domain
    a, b = -L * np.sqrt(tau), L * np.sqrt(tau)
    
    # Summation from k=0 to k=N-1
    k = np.linspace(0, N-1, N).reshape(-1, 1)
    u = k * np.pi / (b - a)
    
    H_k = call_put_coefficients(cp, a, b, k)

    mat = np.exp(1j * np.outer((x0 - a) , u))
    temp = cf(u) * H_k
    temp[0] = 0.5 * temp[0]
    value = np.exp(-r * tau) * K * np.real(mat.dot(temp))

    return value


def call_put_coefficients(cp, a, b, k):
    c, d = (0, b) if cp == 1 else (a, 0)
    coef = chi_psi(a, b, c, d, k)
    
    chi, psi = coef['chi'], coef['psi']
    prefactor = 2 / (b - a)
    
    return prefactor * (chi - psi) if cp == 1 else prefactor * (-chi + psi)


@jit(nopython=True)
def chi_psi(a, b, c, d, k):
    psi = np.sin(k * np.pi * (d - a) / (b - a)) - np.sin(k * np.pi * (c - a)/ (b - a))
    psi[1:] = psi[1:] * (b - a) / (k[1:] * np.pi)
    psi[0] = d - c
    
    chi = 1.0 / (1.0 + np.power((k * np.pi / (b - a)) , 2.0))
    expr1 = np.cos(k * np.pi * (d - a)/(b - a)) * np.exp(d) - np.cos(k * np.pi  * (c - a) / (b - a)) * np.exp(c)
    expr2 = k * np.pi / (b - a) * np.sin(k * np.pi * (d - a) / (b - a)) - k * np.pi / (b - a) * np.sin(k * np.pi * (c - a) / (b - a)) * np.exp(c)
    chi = chi * (expr1 + expr2)
    
    return {"chi": chi, "psi": psi}


def bs_call_option_price(cp, S_0, K, σ, tau, r):
    K = np.array(K).reshape(-1, 1)
    d1 = (np.log(S_0 / K) + (r + 0.5 * σ ** 2) * tau) / (σ * np.sqrt(tau))
    d2 = d1 - σ * np.sqrt(tau)

    if cp == 1:
        return norm.cdf(d1) * S_0 - norm.cdf(d2) * K * np.exp(-r * tau)
    elif cp == -1:
        return norm.cdf(-d2) * K * np.exp(-r * tau) - norm.cdf(-d1) * S_0


# Implied volatility method
def implied_volatility(cp, market_price, K, T, S_0, r):
    func = lambda sigma: (bs_call_option_price(cp, S_0, K, sigma, T, r) - market_price) ** 1.0
    return optimize.newton(func, 0.7, tol=1e-5)


def chf_heston_model(r, tau, kappa, gamma, vbar, v0, rho):
    D1 = lambda u: np.sqrt((kappa - gamma * rho * 1j * u) ** 2 + (u ** 2 + 1j * u) * gamma ** 2)
    g = lambda u: (kappa - gamma * rho * 1j * u - D1(u)) / (kappa - gamma * rho * 1j * u + D1(u))
    C = lambda u: (1 - np.exp(-D1(u) * tau)) / (gamma ** 2 * (1 - g(u) * np.exp(-D1(u) * tau)))
    A = lambda u: r * 1j * u * tau + kappa * vbar * tau / gamma ** 2 * (kappa - gamma * rho * 1j * u - D1(u)) \
                 - 2 * kappa * vbar / gamma ** 2 * np.log((1 - g(u) * np.exp(-D1(u) * tau)) / (1 - g(u)))
    return lambda u: np.exp(A(u) + C(u) * v0)



cp = 1
S0 = 100
r = 0.00
tau = 1

K = np.linspace(50,150,10000)
K = np.array(K).reshape([len(K),1])

# COS method settings
L = 3

kappa = 1.5768
gamma = 0.5751
vbar = 0.0398
rho =-0.5711
v0 = 0.0175

# Compute ChF for the Heston model
cf = chf_heston_model(r,tau,kappa,gamma,vbar,v0,rho)

# The COS method -- Reference value
N = 5000
valCOSRef = call_put_option_price_cos_method(cf, cp, S0, r, tau, K, N, L)

plt.figure(1)
plt.plot(K,valCOSRef,'-k')
plt.xlabel("strike, K")
plt.ylabel("Option Price")
plt.grid()

# For different numbers of integration terms

N = [32,64,96,128,160,400]
maxError = []
for n in N:
    valCOS = call_put_option_price_cos_method(cf, cp, S0, r, tau, K, n, L)
    errorMax = np.max(np.abs(valCOS - valCOSRef))
    maxError.append(errorMax)
    print('maximum error for n ={0} is {1}'.format(n,errorMax))
        




# -*- coding: utf-8 -*-
import datetime as dt
from functools import reduce
import os
import sys

import holidays
import numpy as np
import pandas as pd
from dateutil.relativedelta import relativedelta

years = range(1990,2100)

LOCALE_HOLIDAY = dict(sorted({
    'AE-DUBAI': holidays.AE(categories=['public'], years=years),
    'AR-BUENOS_AIRES': holidays.AR(categories=['public'], years=years),
    'AU-SYDNEY': holidays.AU(subdiv='NSW', categories=['public', 'bank'], years=years),
    'AU-MELBOURNE': holidays.AU(subdiv='VIC', categories=['public', 'bank'], years=years),
    'BR-SAO_PAULO': holidays.BR(subdiv='SP', categories=['public'], years=years),
    'CA-TORONTO': holidays.CA(subdiv='ON', categories=['public'], years=years),
    'CH-ZURICH': holidays.CH(subdiv='ZH', categories=['public'], years=years),
    'CL-SANTIAGO': holidays.CL(subdiv='RM', categories=['public', 'bank'], years=years),
    'CN-SHANGHAI': holidays.CN(categories=['public'], years=years),
    'CO-BOGOTA': holidays.CO(categories=['public'], years=years),
    'CZ-PRAGUE': holidays.CZ(categories=['public'], years=years),
    'DK-COPENHAGEN': holidays.DK(categories=['public'], years=years),
    'DE-FRANKFURT': holidays.DE(subdiv='HH', categories=['public'], years=years),
    'ES-MADRID': holidays.ES(subdiv='MD', categories=['public'], years=years),
    'FR-PARIS': holidays.FR(categories=['public'], years=years),
    'FI-HELSINKI': holidays.FI(categories=['public'], years=years),
    'HK-HONG_KONG': holidays.HK(categories=['public'], years=years),
    'HU-BUDAPEST': holidays.HU(categories=['public'], years=years),
    'ID-JAKARTA': holidays.ID(categories=['public'], years=years),
    'IN-MUMBAI': holidays.IN(subdiv='MH', categories=['public'], years=years),
    'IL-TEL_AVIV': holidays.IL(categories=['public'], years=years),
    'IS-REYKJAVIK': holidays.IS(categories=['public'], years=years),
    'MY-KUALA_LUMPUR': holidays.MY(subdiv='KUL', categories=['public'], years=years),
    'MX-MEXICO_CITY': holidays.MX(categories=['public'], years=years),
    'NO-OSLO': holidays.NO(categories=['public'], years=years),
    'NZ-AUCKLAND': holidays.NZ(subdiv='AUK', categories=['public'], years=years),
    'PL-WARSAW': holidays.PL(categories=['public'], years=years),
    'PH-MANILA': holidays.PH(categories=['public'], years=years),
    'RU-MOSCOW': holidays.RU(categories=['public'], years=years),
    'SA-RIYADH': holidays.SA(categories=['public'], years=years),
    'SE-STOCKHOLM': holidays.SE(categories=['public'], years=years),
    'SG-SINGAPORE': holidays.SG(categories=['public'], years=years),
    'TH-BANKCOK': holidays.TH(categories=['public', 'bank'], years=years),
    'TR-ISTANBUL': holidays.TR(categories=['public'], years=years),
    'TW-TAPEI': holidays.TW(categories=['public'], years=years),
    'UK-LONDON': holidays.UK(subdiv='ENG', categories=['public'], years=years),
    'US-NEW_YORK': holidays.US(subdiv='NY', categories=['public'], years=years),
    'VN-HO_CHI_MINH_CITY': holidays.VN(categories=['public'], years=years),
    'ZA-JOHANNESBURG': holidays.ZA(categories=['public'], years=years),
    'European_Central_Bank': holidays.ECB(categories=['public'], years=years),
    'ICE_Futures_Europe': holidays.IFEU(categories=['public'], years=years),
    'New_York_Stock_Exchange': holidays.XNYS(categories=['public'], years=years),
}.items()))


CCY_HOLIDAY = dict(sorted({
    'AED': holidays.AE(categories=['public'], years=years),
    'ARS': holidays.AR(categories=['public'], years=years),
    'AUD': holidays.AU(subdiv='NSW', categories=['public', 'bank'], years=years),
    'BRL': holidays.BR(subdiv='SP', categories=['public'], years=years),
    'CAD': holidays.CA(subdiv='ON', categories=['public'], years=years),
    'CHF': holidays.CH(subdiv='ZH', categories=['public'], years=years),
    'CLP': holidays.CL(subdiv='RM', categories=['public', 'bank'], years=years),
    'CNY': holidays.CN(categories=['public'], years=years),
    'COP': holidays.CO(categories=['public'], years=years),
    'CZK': holidays.CZ(categories=['public'], years=years),
    'DKK': holidays.DK(categories=['public'], years=years),
    'HKD': holidays.HK(categories=['public'], years=years),
    'HUF': holidays.HU(categories=['public'], years=years),
    'IDR': holidays.ID(categories=['public'], years=years),
    'INR': holidays.IN(subdiv='MH', categories=['public'], years=years),
    'ILS': holidays.IL(categories=['public'], years=years),
    'ISK': holidays.IS(categories=['public'], years=years),
    'MYR': holidays.MY(subdiv='KUL', categories=['public'], years=years),
    'MXN': holidays.MX(categories=['public'], years=years),
    'NOK': holidays.NO(categories=['public'], years=years),
    'NZD': holidays.NZ(subdiv='AUK', categories=['public'], years=years),
    'PLN': holidays.PL(categories=['public'], years=years),
    'PHP': holidays.PH(categories=['public'], years=years),
    'RUB': holidays.RU(categories=['public'], years=years),
    'SAR': holidays.SA(categories=['public'], years=years),
    'SEK': holidays.SE(categories=['public'], years=years),
    'SGD': holidays.SG(categories=['public'], years=years),
    'THB': holidays.TH(categories=['public', 'bank'], years=years),
    'TRY': holidays.TR(categories=['public'], years=years),
    'TWD': holidays.TW(categories=['public'], years=years),
    'GBP': holidays.UK(subdiv='ENG', categories=['public'], years=years),
    'USD': holidays.US(subdiv='NY', categories=['public'], years=years),
    'VND': holidays.VN(categories=['public'], years=years),
    'ZAR': holidays.ZA(categories=['public'], years=years),
    'EUR': holidays.ECB(categories=['public'], years=years),
}.items()))

#%%

def get_holidays_obj(ccy):
    ccy_uppercase = ccy.upper()
    if ccy_uppercase in CCY_HOLIDAY.keys():
        return CCY_HOLIDAY[ccy_uppercase]
    else:
        raise ValueError(f"Holidays not setup for {ccy_uppercase}")


def convert_weekend_to_weekmask(weekend_set):
    "For converting the holidays.weekend attribute to a valid weekmask for numpy.busdaycalendar"
    # Initialize a list of ones representing Monday to Sunday (working days)
    weekmask = [1, 1, 1, 1, 1, 1, 1]
    
    # Loop through the weekend days and set corresponding indexes in the weekmask to 0
    for weekend_day in weekend_set:
        weekmask[weekend_day] = 0
    
    # Return the weekmask as a string (optional, can also return as a list if needed)
    return ''.join(map(str, weekmask))


def get_busdaycalendar(ccys) -> np.busdaycalendar:
    """
    Create a calendar which has the holidays and business days of the currency inputs.
    """

    if ccys is None:
        return np.busdaycalendar()

    if type(ccys) is str:
        ccys = [ccys]
    elif type(ccys) is list:
        ccys = list(set(ccys))
        ccys = [ccy.upper() for ccy in ccys]

    holiday_objs = [get_holidays_obj(ccy) for ccy in ccys]
    combined_weekend_union = reduce(lambda x, y: x | y.weekend, holiday_objs, set())
    weekmasks = convert_weekend_to_weekmask(combined_weekend_union)

    # Flattan the list of lists
    holidays = [h for holiday_list in holiday_objs for h in holiday_list]

    return np.busdaycalendar(weekmask=weekmasks, holidays=holidays)  # .values.astype('datetime64[D]'))


# %% Create code for static holidays definition

if __name__ == "__main__":
    # Get the directory of the current file
    current_dir = os.path.dirname(os.path.abspath(__file__))

    # Add the directory to the system path
    sys.path.append(current_dir)
    df = pd.read_excel(current_dir + "\\calendar_map.xlsx", usecols=range(6))
    holiday_dict = {}
    holiday_dict_str = "LOCALE_HOLIDAY = dict(sorted({\n"
    ccy_holiday_dict = {}
    ccy_holiday_dict_str = "CCY_HOLIDAY = dict(sorted({\n"
    years=range(1990,2100)

    for i, row in df.iterrows():
        eval_str = None
        ccy = row["ISOCurrencyCode"]
        ui_string = row["UIString"]
        code = row["Code"]
        subdivision = row["Subdivision"]

        try:
            categories = [holidays.PUBLIC, holidays.BANK]
            if pd.isna(row["Subdivision"]):
                eval_str = f"holidays.{code}(categories={categories}, years=years)"

            else:
                eval_str = (
                    f"holidays.{code}(subdiv='{subdivision}', categories={categories}, years=years)"
                )
            holiday_obj = eval(eval_str)

        except ValueError:
            categories = [holidays.PUBLIC]
            if pd.isna(row["Subdivision"]):
                eval_str = f"holidays.{code}(categories={categories}, years=years)"
            else:
                eval_str = (
                    f"holidays.{code}(subdiv='{subdivision}', categories={categories}, years=years)"
                )
            holiday_obj = eval(eval_str)

        holiday_dict[ui_string] = holiday_obj
        holiday_dict_str += f"    '{ui_string}': {eval_str},\n"

        if row["CurrencyPrimaryCalendar"]:
            ccy_holiday_dict[ccy] = holiday_obj
            ccy_holiday_dict_str += f"    '{ccy}': {eval_str},\n"

    holiday_dict_str += "}.items()))"
    ccy_holiday_dict_str += "}.items()))"

    print(holiday_dict_str)
    print("\n")
    print(ccy_holiday_dict_str)



# -*- coding: utf-8 -*-


if __name__ == "__main__":
    import os
    import pathlib
    os.chdir(pathlib.Path(__file__).parent.parent.parent.parent.resolve())     
    print('__main__ - current working directory:', os.getcwd())

from frm.frm.schedule.daycounter import DayCounter, VALID_DAY_COUNT_BASIS
from frm.frm.schedule.schedule import payment_schedule, VALID_DAY_ROLL, VALID_PAYMENT_TYPE, VALID_PAYMENT_FREQUENCY, VALID_STUB, VALID_STUB_GENERAL, VALID_ROLL_CONVENTION
from frm.frm.schedule.business_day_calendar import get_calendar
from frm.frm.schedule.tenor import calc_tenor_date
from frm.frm.market_data.iban_ccys import VALID_CCYS

import numpy as np
import pandas as pd
from scipy.optimize import fsolve
from scipy.stats import norm

from dataclasses import dataclass, field, InitVar 
from typing import Literal, Optional

VALID_EXCHANGE_NOTIONALS = Literal['start','end','both','neither']

@dataclass
class Leg:
    "Generic class to cover pricing of all types of swap legs and bonds"
    
    # Tenor definitions
    # (i) Specifically define the effective/maturity date of the instrument
    effective_date: Optional[pd.Timestamp]=None
    maturity_date: Optional[pd.Timestamp]=None
    # (ii) Alternatively define the tenor and transcation date/trade date/pricing date 
    #      and the effective_date and maturity_date will be computed
    tenor: Optional[pd.DateOffset]=None, 
    transaction_date: Optional[pd.Timestamp]=None
    forward_starting: int=2 # number of business days after the transaction date the instrument starts
    
    transaction_price: Optional[float]=100.0 # As a % of par. 100.0 is the par pricing
    
    # Coupon definitions
    pay_rec: Optional[Literal['pay','rec']]=None
    leg_type: Optional[Literal['float','fixed']]=None
    fixed_rate: Optional[float]=None
    float_spread: Optional[float]=None
    cpn_cap: Optional[float]=None
    cpn_floor: Optional[float]=None
    day_count_basis: InitVar[VALID_DAY_COUNT_BASIS] = 'ACT/ACT'
    
    # Payment schedule definitions
    payment_frequency: VALID_PAYMENT_FREQUENCY='quarterly'
    roll_convention: VALID_ROLL_CONVENTION='modifiedfollowing'
    day_roll: Optional[VALID_DAY_ROLL]=None
    stub: VALID_STUB_GENERAL='first_short'
    first_stub: VALID_STUB='short'
    last_stub: Optional[VALID_STUB]=None
    first_cpn_end_date: Optional[pd.Timestamp]=None
    last_cpn_start_date: Optional[pd.Timestamp]=None
    payment_type: VALID_PAYMENT_TYPE='in_arrears'
    payment_delay: int=0
    fixing_days_ahead: int=0
    currency_holidays=None
    city_holidays=None
    holiday_calendar: InitVar[Optional[np.busdaycalendar]]=None
    
    # Notional and currencies definitions
    notional: float=100e6
    notional_currency: Optional[str]=None
    settlement_currency: Optional[str]=None
    exchange_notionals: VALID_EXCHANGE_NOTIONALS='both'
    MTM_notional_currency: Optional[str]=None
    
    # Pricing curves
    dsc_crv: Optional['ZeroCurve']=None
    fwd_crv: Optional['ZeroCurve']=None
    
    # Non initialisation arguments
    daycounter: DayCounter = field(init=False)
    
    
    def __post_init__(self, day_count_basis, holiday_calendar):
             
                
        if self.forward_starting is None: self.forward_starting = 2
        if self.roll_convention is None: self.roll_convention='modifiedfollowing'
        if self.stub is None: self.stub = 'first_short'
        if self.payment_type is None: self.payment_type ='in_arrears'
        if self.payment_delay is None: self.payment_delay=0
        if self.fixing_days_ahead is None: self.fixing_days_ahead=2
        if self.notional is None: self.notional=100e6
        if self.exchange_notionals is None: self.exchange_notionals = 'both'
        
        assert bool(self.effective_date is not None and self.maturity_date is not None) \
            != bool(self.tenor is not None and self.transaction_date is not None)
        assert self.fixed_rate is None or self.float_spread is None
        assert self.payment_frequency is not None
        assert self.notional_currency is not None
        assert self.notional > 0
        
        if self.settlement_currency is None: self.settlement_currency = self.notional_currency
        if self.currency_holidays is None: self.currency_holidays = [self.notional_currency, self.settlement_currency]
        if holiday_calendar is None: 
            holiday_calendar = get_calendar(self.currency_holidays, self.city_holidays) 
        if self.effective_date is None: 
            self.effective_date = np.busday_offset(pd.DatetimeIndex([self.transaction_date + pd.DateOffset(days=self.forward_starting)]).values.astype('datetime64[D]'), offsets=0, roll=self.roll_convention, busdaycal=holiday_calendar)[0]
        if self.maturity_date is None:
            tenor_date, _, _ = calc_tenor_date(self.transaction_date, self.tenor, curve_ccy=self.notional_currency, holiday_calendar=holiday_calendar)
            self.maturity_date = tenor_date
            
        
            #self.maturity_date = np.busday_offset(pd.DatetimeIndex([self.effective_date + self.tenor]).values.astype('datetime64[D]'), offsets=0, roll='following', busdaycal=holiday_calendar)[0]

        self.daycounter = DayCounter(day_count_basis)
                
        if self.fixed_rate is not None: self.leg_type = 'fixed'
        if self.float_spread is not None: self.leg_type = 'float'

        self.schedule = payment_schedule(start_date=self.effective_date,
                                         end_date=self.maturity_date,
                                         payment_freq=self.payment_frequency,
                                         roll_convention=self.roll_convention,
                                         day_roll=self.day_roll,
                                         stub=self.stub,
                                         first_stub=self.first_stub,
                                         last_stub=self.last_stub,
                                         first_cpn_end_date=self.first_cpn_end_date,
                                         last_cpn_start_date=self.last_cpn_start_date,
                                         payment_type=self.payment_type,
                                         payment_delay=self.payment_delay,
                                         add_fixing_dates=self.leg_type == 'float',
                                         add_initial_exchange_period=self.leg_type in {'both','start'},
                                         fixing_days_ahead=self.fixing_days_ahead,
                                         currency_holidays=self.currency_holidays,
                                         city_holidays=self.city_holidays,
                                         holiday_cal=self.holiday_calendar)
        self.schedule['period_length'] = self.daycounter.year_fraction(self.schedule['period_start'], self.schedule['period_end'])        

    def __price_helper(self, 
                       schedule: pd.DataFrame,
                       dsc_crv=None,
                       fwd_crv=None) -> (float, pd.DataFrame):
        
        if dsc_crv is None: dsc_crv = self.dsc_crv
        if fwd_crv is None: fwd_crv = self.fwd_crv
        if schedule is None: schedule = self.schedule
        
        if self.leg_type == 'float':
            schedule['forward_reset'] = fwd_crv.forward_rate(schedule['fixing_period_start'],schedule['fixing_period_end'])
            schedule['coupon_rate'] = schedule['forward_reset'] + schedule['spread']
        schedule['coupon_cashflow_amount'] = schedule['notional'] * schedule['coupon_rate'] * schedule['period_length']            
        schedule['net_cashflow_amount'] = schedule['principal_cashflow_amount'] + schedule['coupon_cashflow_amount']
        
        if self.notional_currency == self.settlement_currency:
            schedule['notional_per_settle_fx_forward_rate'] = 1.0
        else:
            raise ValueError('to be built')     
        
        schedule['net_settlement_amount'] = schedule['net_cashflow_amount'] / schedule['notional_per_settle_fx_forward_rate']
        schedule['discount_factor'] = dsc_crv.discount_factor(schedule['payment_date'])            
        schedule['net_settlement_value'] = schedule['net_settlement_amount'] * schedule['discount_factor']
        
        schedule['settlement_currency'] = self.settlement_currency
        price = schedule['net_settlement_value'].sum()
        if self.pay_rec == 'pay':
            price = -price
        
        return price, schedule

           
    def price(self, 
              calc_PV01: bool=False, 
              fixed_rate_overide: Optional[float]=None, 
              float_spread_overide: Optional[float]=None) -> dict:
                
        # Set the notional amounts 
        # If CCIRS was MTM calculate the MTM amounts here
        schedule = self.schedule.copy()
        schedule['notional'] = self.notional
        schedule['notional_currency'] = self.notional_currency

        # Specify notional cashflow payments      
        schedule['principal_cashflow_amount'] = 0.0
        if self.exchange_notionals in {'both','start'}:
            schedule.loc[0,'principal_cashflow_amount'] = - self.notional
        if self.exchange_notionals in {'both','end'}:
            schedule.loc[len(schedule)-1,'principal_cashflow_amount'] = self.notional
        
        if self.leg_type == 'fixed':
            if fixed_rate_overide is None: 
                schedule['coupon_rate'] = self.fixed_rate
            else:
                schedule['coupon_rate'] = fixed_rate_overide[0]
        if self.leg_type == 'float':
            if float_spread_overide is None: 
                schedule['spread'] = self.float_spread
            else:
                schedule['spread'] = float_spread_overide[0]
        
        pricing = {}
        pricing['price'], pricing['cashflows'] = self.__price_helper(schedule=schedule)
        
        if calc_PV01: 
            pricing['price_discount_shift'], _ = self.__price_helper(schedule=schedule, dsc_crv=self.dsc_crv.flat_shift())
            pricing['price_forward_shift'], _ = self.__price_helper(schedule=schedule, fwd_crv=self.fwd_crv.flat_shift())
            pricing['price_discount_and_forward_shift'], _ = self.__price_helper(schedule=schedule,
                                                                                         dsc_crv=self.dsc_crv.flat_shift(), 
                                                                                         fwd_crv=self.fwd_crv.flat_shift())
        return pricing
                 

    def solver(self,
               solve_price: Optional[float]=None) -> (float, str):
        
        if solve_price is None:
            if self.pay_rec == 'pay':
                solve_price = -self.notional
            elif self.pay_rec == 'rec':
                solve_price = self.notional
            
        if self.leg_type == 'fixed':
            initial_rate = self.fixed_rate
            def solver_helper(fixed_rate: [float], leg:Leg, solve_price: float) -> float:
                return leg.price(fixed_rate_overide=fixed_rate, calc_PV01=False)['price'] - solve_price
        elif self.leg_type == 'float':
            initial_rate = self.float_spread
            def solver_helper(float_spread: [float], leg:Leg, solve_price: float) -> float:
                return leg.price(float_spread_overide=float_spread, calc_PV01=False)['price'] - solve_price
                
        x, infodict, ier, msg = fsolve(solver_helper, [initial_rate], args=(self, solve_price), xtol=1, full_output=True)
        return x[0], msg
        
                
    def set_discount_curve(self, zero_crv):   
        self.dsc_crv = zero_crv

    def set_forward_curve(self, zero_crv):   
        self.fwd_crv = zero_crv
             
    def price_cap_floor(self, d1, d2, K, σ):
        
        T = self.daycounter(d1,d2)
        F = self.fwd_crv.forward_rate(d1,d2) 
        
        d1 = (np.log(F/K) + (0.5 * σ**2 * T)) / (σ*np.sqrt(T))
        d2 = d1 - σ*np.sqrt(T)
        
        bought_cap = F * norm.cdf(d1) - K * norm.cdf(d2)
        bought_put = K * norm.cdf(d2) - F * norm.cdf(d1)
        
        
        
        
        
                
                
                
                
        
        
               


# -*- coding: utf-8 -*-


import os
import pathlib

os.chdir(pathlib.Path(__file__).parent.parent.resolve()) 

from schedule.daycounter import DayCounter, VALID_DAY_COUNT_BASIS
from schedule.schedule import payment_schedule, VALID_DAY_ROLL, VALID_PAYMENT_TYPE, VALID_PAYMENT_FREQUENCY, VALID_STUB, VALID_STUB_GENERAL, VALID_ROLL_CONVENTION
from calendars.calendar import get_calendar, VALID_CITY_HOLIDAYS, VALID_CURRENCY_HOLIDAYS
from market_data.iban_currencies import VALID_CURRENCIES
from instruments.leg import Leg

import datetime as dt
import numpy as np
import pandas as pd
from scipy.optimize import fsolve
from dataclasses import dataclass, InitVar
from typing import Optional, Literal

                      
# MARKET_ISSUE = Literal['DOMESTIC''EURO_MTN']
# VALID_DEBT_CLASSES = Literal['first_lien', 'senior_secured', 'subordinated']


@dataclass
class Bond():
    
    # ************************** Pricing attributes ***************************
    # The pricing attributes are used by the Leg class
    
    # Tenor definitions
    # (i) Specifically define the effective/maturity date of the instrument
    effective_date: Optional[pd.Timestamp]=None
    maturity_date: Optional[pd.Timestamp]=None
    # (ii) Alternatively define the tenor and transcation date/trade date/pricing date 
    #      and the effective_date and maturity_date will be computed
    tenor: Optional[pd.DateOffset]=None, 
    transaction_date: Optional[pd.Timestamp]=None
    forward_starting: int=0 # number of days after the transaction date the instrument starts
    
    transaction_price: Optional[float]=100.0 # As a % of par. 100.0 is the par pricing     
    
    # Coupon definitions
    pay_rec: Optional[Literal['pay','rec']]='rec'
    leg_type: Optional[Literal['float','fixed']]=None
    fixed_rate: Optional[float]=None
    float_spread: Optional[float]=None
    cpn_cap: Optional[float]=None
    cpn_floor: Optional[float]=None
    day_count_basis: [VALID_DAY_COUNT_BASIS] = '30/360'
    
    # Payment schedule definitions
    payment_frequency: VALID_PAYMENT_FREQUENCY='semiannually'
    roll_convention: VALID_ROLL_CONVENTION='modifiedfollowing'
    day_roll: Optional[VALID_DAY_ROLL]=None
    stub: VALID_STUB_GENERAL='first_short'
    first_stub: VALID_STUB='short'
    last_stub: Optional[VALID_STUB]=None
    first_cpn_end_date: Optional[pd.Timestamp]=None
    last_cpn_start_date: Optional[pd.Timestamp]=None
    payment_type: VALID_PAYMENT_TYPE='in_arrears'
    payment_delay: int=0
    fixing_delay: int=0
    currency_holidays: Optional[VALID_CURRENCY_HOLIDAYS]=None
    city_holidays: Optional[VALID_CITY_HOLIDAYS]=None
    holiday_calendar: [Optional[np.busdaycalendar]]=None
    
    # Notional and currencies definitions
    notional: float=100e6
    currency: Optional[VALID_CURRENCIES]=None
    
    # Accrued interest calculation
    calculation_type: Optional[int]=None
    settlement_delay: Optional[int]=None # T+1 T+2 T+3 ...
    ex_dividend_days: Optional[int]=None # number of business days prior to the coupon payment date you must own the bond to get the bond
        
    # Pricing curves
    dsc_crv: Optional['ZeroCurve']=None
    fwd_crv: Optional['ZeroCurve']=None
    
    # ************************ Description attributes *************************
    # amount_issued: Optional[float]=None
    # min_piece: Optional[float]=None
    # min_increment: Optional[float]=None

    # issuance_country: Optional[str]=None
    # market_issue: Optional[MARKET_ISSUE]=None
    # debt_class: Optional[VALID_DEBT_CLASSES]=None
    # bond_ratings: Optional[dict]=None
    
    # # Identifiers
    # ISIN: Optional[str]=None
    # CUSIP: Optional[str]=None
    # SEDOL: Optional[str]=None
    # FIGI: Optional[str]=None
    

    def __post_init__(self):
    
        if self.fixed_rate is not None and self.float_spread is None: self.leg_type = 'fixed'
        if self.fixed_rate is None and self.float_spread is not None: self.leg_type = 'float'
    
        self.leg = Leg(
            # Tenor definitions
            effective_date=self.effective_date,
            maturity_date=self.maturity_date, 
            tenor=self.tenor,
            transaction_date=self.transaction_date,
            forward_starting=self.forward_starting,
            transaction_price=self.transaction_price,
            # Payment schedule definitions
            payment_frequency=self.payment_frequency,
            roll_convention=self.roll_convention,
            day_roll=self.day_roll,
            stub=self.stub,
            first_stub=self.first_stub,
            last_stub=self.last_stub,
            first_cpn_end_date=self.first_cpn_end_date,
            last_cpn_start_date=self.last_cpn_start_date,
            payment_type=self.payment_type, 
            payment_delay=self.payment_delay,
            fixing_delay=self.fixing_delay,
            currency_holidays=self.currency_holidays,
            city_holidays=self.city_holidays,
            holiday_calendar=self.holiday_calendar,
            # Coupon definitions
            pay_rec='pay' if self.buy_sell == 'sell' else 'rec',
            leg_type=self.leg_type,
            fixed_rate=self.fixed_rate,
            float_spread=self.float_spread,
            cpn_cap=self.cpn_cap,
            cpn_floor=self.cpn_floor,
            day_count_basis=self.day_count_basis,
            # Notional and currencies definitions
            notional=self.notional,
            notional_currency=self.currency,
            exchange_notionals='end',
            # Pricing curves
            dsc_crv=self.dsc_crv,
            fwd_crv=self.fwd_crv)

    def price(self, 
              calc_PV01: bool=False, 
              fixed_rate_overide: Optional[float]=None, 
              float_spread_overide: Optional[float]=None) -> dict:
                
        pricing = self.leg.price(calc_PV01, fixed_rate_overide, float_spread_overide) 
        
        return pricing
                 

    def solver(self,
               solve_price: Optional[float]=None) -> (float, str):
        
        x, msg = self.leg.solver(solve_price)
        
        return x, msg
           
        


# -*- coding: utf-8 -*

import numpy as np
from scipy.stats import norm


if __name__ == "__main__":
    import os
    import pathlib
    import sys
    
    os.chdir(pathlib.Path(__file__).parent.parent.parent.parent.resolve()) 
    sys.path.append(os.getcwd())
    print('__main__ - current working directory:', os.getcwd())


#%% Test COS method function with the normal distribution

def cos_method(p, cf, dt, a, b, N=160):
    """
    PROBLEM STATEMENT: The charactheristic function (or moment generating function, or
    Laplace transform, from which to retrieve it) of a certain random variable X is known. 
    The objective is approximating f and F, respectively PDF and CDF of X.

    Parameters
    ----------
    p : TYPE
        point at which to evaluate the CDF (F) of f - must be in [a,b].
    cf : TYPE
        characteristic function provided as a function handle where the only variable is the point at which to evaluate it. 
    N : int
        Number of sums, per [1], N=160 the COS method quickly priced options with high accuracy
    dt : float
        Step size for approximating the integral
    a : float
        Lower integration bound
    b : float
        Upper integration bound

    Returns
    -------
    Fp : float
        CDF evaluated at p
    F : float
        CDF evaluated on pts
    f : float
        PDF evaluated on pts
    I : float
        Integral of the PDF over [a,b]: 
    pts : float
        points over which the CDF and PDF are sampled: a:dt:pts (and point p)    
        
        
    References:
    [1] Fang, Fang & Oosterlee, Cornelis. (2008). A Novel Pricing Method for European Options Based on Fourier-Cosine Series Expansions. SIAM J. Scientific Computing. 31. 826-848. 10.1137/080718061. 
    """
    
    assert dt > 0.0    
    assert (a + dt) <= b
    assert p >= a
    assert p <= b
    
    pts = np.sort(np.unique(np.concatenate(([p], np.arange(a, b + dt, dt))))) # Array from [a, a+dt, a+2*dt, ..., b]
    dt_pts = np.append(np.diff(pts), dt)
    k = np.arange(N + 1) # Array from [0, 1, 2, ... N, N+1]

    w = np.append(0.5, np.ones(N))
    
    Fk = 2 / (b-a) * np.real(cf(k * np.pi / (b-a)) * np.exp(-1j * k * np.pi * a / (b-a))) # Equation 8 in [1] 
    C = lambda x: np.cos(k * np.pi * (x - a) / (b-a)) # Inner term of equation 11 in [1] 
    PDF_pts = np.array([np.sum(w * (Fk * C(x))) for x in pts]) # Equation 11 in [1]

    integral_of_a_to_b = np.sum(PDF_pts * dt_pts)
    if integral_of_a_to_b < norm.cdf(5):
        raise ValueError('The integral of the PDF over [a,b] is not close enough to 1. Please increase the range of [a,b]')
    
    CDF_pts = np.cumsum(PDF_pts * dt_pts)
    CDF_p = CDF_pts[np.where(pts == p)[0][0]]

    return CDF_p, CDF_pts, PDF_pts, integral_of_a_to_b, pts


def get_cos_truncation_range(model, L, model_param):
    """
    # This function returns the truncation range of the characteristic function 
    # Defined for COS method approximations of stochastic models as detailed in Appendix of Fang & Oosterlee (2008), in Table 11, page 21/21.
    # Note: c1, c2, ... cn are the 1st, 2nd, .. nth cumulant of ln(ST/K)
    # References:
    # [1] Fang, Fang & Oosterlee, Cornelis. (2008). A Novel Pricing Method for European Options Based on Fourier-Cosine Series Expansions. SIAM J. Scientific Computing. 31. 826-848. 10.1137/080718061. 
        
    Parameters
    ----------
    model : str
        the type of stochastic model (geomentric_brownion_motion, heston, VG, CGMY).
    model_param : dict
        defining paramaters of the stochastic model

    Returns
    -------
    a : float
        lower truncation value.
    b : float
        upper truncation value.

    """
    
    # From Table 11 in the Appendix A of [1]
    if model == 'geomentric_brownion_motion':
        # GBM is defined by two parameters:
        # mu: the drift
        # σ: the volatility
        mu,σ,tau = model_param
        
        c1 = mu * tau
        c2 = (σ**2) * tau
        c4 = 0
     
    elif model == 'heston':
        # Remap the parameters per the symbols used in [1] so comparison to the formulae on page 21 of 21 is easier
        tau = model_param['tau'] # time-to-expiry
        mu = model_param['mu'] # drift mu. mu = r-q. For FXO mu = r_DOM - r_FOR
        u0 = model_param['var0'] # Initial variance.
        η  = model_param['vv'] # Volatility of volatility.
        λ = model_param['kappa'] # rate of mean reversion to the long-run variance
        u_bar = model_param['theta'] # Long-run variance
        rho = model_param['rho'] # Correlation.
                        
        c1 = mu * tau + (1- np.exp(-1 * λ * tau)) * (u_bar - u0) / (2 * λ) - 0.5 * u_bar * tau
        
        c2 = (1 / (8 * λ**3)) * (
            (η * tau * λ * np.exp(-λ * tau) * (u0 - u_bar) * (8 * λ * rho - 4 * η))
            + λ * rho * η * (1 - np.exp(-λ * tau)) * (16 * u_bar - 8 * u0)
            + 2 * u_bar * λ * tau * (-4 * λ * rho * η + η**2 + 4 * λ**2)
            + η**2 * ((u_bar - 2 * u0) * np.exp(-2 * λ * tau) + u_bar * (6 * np.exp(-λ * tau) - 7) + 2 * u0)
            + 8 * λ**2 * (u0 - u_bar) * (1 - np.exp(-λ * tau))
        )
                
        c4 = 0
        
        #print("SF: c1,c2", c1, c2)
          
    elif model == 'VG':
        # Detailed in paper but not priority for this project
        pass
    elif model == 'CGMY':
        # Detailed in paper but not priority for this project
        pass    
        
    a = c1 - L * np.sqrt(np.abs(c2) + np.sqrt(np.abs(c4))) # Per Equantion 49 from [1]
    b = c1 + L * np.sqrt(np.abs(c2) + np.sqrt(np.abs(c4))) # Per Equantion 49 from [1]

    return a,b










# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd
from scipy.stats import norm
from scipy.optimize import root_scalar, newton

def gk_price(S0: float, 
             tau: float, 
             r_f: float,  
             r_d: float,               
             cp: int,
             K: float,
             σ: float, 
             F: float = None,
             analytical_greeks_flag: bool=False,
             numerical_greeks_flag: bool=False,
             intrinsic_time_split_flag: bool=False
             ) -> float:
    """
    GarmanKohlhagen European FX option pricing formula.
    - if tau is 0, intrinsic value is returned
    x
    :param S0: FX spot, specified in # of units of domestic per 1 unit of foreign currency
    :param tau: time to np.expiry (in years)
    :param r_d: domestic risk free interest rate (annualised)
    :param r_f: foreign riskless interest rate (annualised)
    :param cp: option type (1 for call option (default), -1 for put option)
    :param K: strike, specified in # of units of term currency per 1 unit of base currency
    :param σ: volatility
    :param F: market forward rate, if None this is calculated under interest rate parity     
    :param analytical_greeks_flag: boolean flag for in/excluding analytically calculated greeks
    :param numerical_greeks_flag: boolean flag for in/excluding numerically calculated greeks
    :return: option price (in the domestic currency, per 1 unit of foreign currency notional), 
             analytical greeks (if analytical_greeks_flag=True),
             numerical greeks (if numerical_greeks_flag=True)
             
    """
  
    S0 = np.atleast_1d(S0).astype(float)
    tau = np.atleast_1d(tau).astype(float)
    r_f = np.atleast_1d(r_f).astype(float)
    r_d = np.atleast_1d(r_d).astype(float)
    cp = np.atleast_1d(cp).astype(float)
    K = np.atleast_1d(K).astype(float)
    σ = np.atleast_1d(σ).astype(float)
    
    # Sensical value checks
    # No >0 check for σ, as when doing numerical solving, need to allow for -ve σ
    assert (S0 > 0.0).all(), S0
    assert (tau >= 0.0).all(), tau # if tau is 0.0, return instrinsic value
    assert np.all(np.isin(cp, [1, -1])), cp
    assert cp.shape == σ.shape
    assert cp.shape == K.shape
    
    results = dict()
    
    if F is not None: 
        F = np.atleast_1d(F).astype(float)
        assert (F > 1e-8).all() 
        
        # Use market forward rate and imply basis-adjusted domestic interest rate        
        r_d_basis_adj = np.log(F / S0) / tau + r_f # from F = S0 * exp((r_d - r_f) * tau)
        d1 = (np.log(F / K) + (0.5 * σ**2) * tau) / (σ * np.sqrt(tau))
        d2 = d1 - σ * np.sqrt(tau)    
        X = cp * np.exp(-r_d_basis_adj * tau) * (F * norm.cdf(cp * d1) - K * norm.cdf(cp * d2))   
        
        # If time-to-maturity is 0.0, set to intrinsic value 
        X[tau==0] = np.maximum(0, cp * (S0 - K))[tau==0]
        results['option_value'] = X
        
        if intrinsic_time_split_flag:
            X_intrinsic = np.full_like(X, np.nan)
            X_time = np.full_like(X, np.nan)            
            X_intrinsic[tau>0] = np.maximum(0, cp * np.exp(-r_d_basis_adj * tau) * (F - K))[tau>0]
            X_intrinsic[tau==0] = X[tau==0]
            X_time = X - X_intrinsic
            results['intrinsic_value'] = X_intrinsic
            results['time_value'] = X_time
    else:
        # Under interest rate parity 
        F = S0 * np.exp((r_d - r_f) * tau)
        d1 = (np.log(S0 / K) + (r_d - r_f + 0.5 * σ**2) * tau) / (σ * np.sqrt(tau)) # Not actually used in the pricing
        d2 = d1 - σ * np.sqrt(tau)    
        X = cp * (S0 * np.exp(-r_f * tau) * norm.cdf(cp * d1) - K * np.exp(-r_d * tau) * norm.cdf(cp * d2))
        
        # If time-to-maturity is 0.0, set to intrinsic value 
        X[tau==0] = np.maximum(0, cp * (S0 - K))[tau==0]
        results['option_value'] = X
        
        if intrinsic_time_split_flag:
            X_intrinsic = np.full_like(X, np.nan)
            X_time = np.full_like(X, np.nan)            
            X_intrinsic[tau>0] = np.maximum(0, cp * (S0 * np.exp(-r_f * tau) - K * np.exp(-r_d * tau)))[tau>0]
            X_intrinsic[tau==0] = X[tau==0]
            X_time = X - X_intrinsic 
            results['intrinsic_value'] = X_intrinsic
            results['time_value'] = X_time            
        
    if not (analytical_greeks_flag or numerical_greeks_flag):
        return results
    else:
        Δ_shift = 1 / 100 # 1% shift
        σ_shift = 1 / 100 # 1% shift
        θ_shift = 1 / 365.25 # 1 Day
        ρ_shift = 1 / 100 # 1% shift
    
        if analytical_greeks_flag:
            analytical_greeks = {}
            
            ######################## 1st Order Greeks #########################
            
            # Delta, Δ, is the change in an option's price for a small change in the underlying assets price.
            # Δ := ∂X/∂S ≈ (X(S_plus) − X(S_minus)) / (S_plus - S_minus)
            # where X is the option price (whose units is DOM),
            # and S0 is the fx spot price (whose units is DOM/FOR)
            # We multiply by S0 so to return the Δ a % applicable to  to the foreign currency notional
            analytical_greeks['spot_delta'] = S0 * cp * np.exp(-r_f * tau) * norm.cdf(cp * d1)
            analytical_greeks['forward_delta'] = S0 * cp * norm.cdf(cp * d1) 
    
            # Vega, ν, is the change in an options price for a small change in the volatility input
            # ν = ∂X/∂σ ≈ (X(σ_plus) − X(σ_minus)) / (σ_plus - σ_minus)
            # In practice, vega is normalised to measure the change in price for a 1% change in the volatility input
            # Hence we have scaled the numerical vega to a 1% change
            analytical_formula = S0 * np.sqrt(tau) * norm.pdf(d1) * np.exp(-r_f * tau) # identical for calls and puts
            analytical_greeks['vega'] = analytical_formula * 0.01 # normalised to 1% change
    
            # Theta, θ, is the change in an options price for a small change in the time to expiry
            # In practice, theta is normalised to measure the change in price, for a 1 day shorter expiry 
            # Theta includes
            # 1. Time decay: the change in the option's value as time moves forward, that is, as the expiry date moves closer, 
            #               i.e., the time value price tomorrow minus the time value price today.
            # 2. Cost of carry: the interest rate sensitivity that causes the value of the portfolio to change as time progresses, 
            #                   e.g., the cost of carry on spots and forwards is included in the theta value.
            analytical_formula = -(S0*np.exp(-r_d*tau)*norm.pdf(d1 * cp)*σ)/(2*np.sqrt(tau)) \
                + r_f * np.exp(-r_f * tau) * S0 * norm.cdf(d1 * cp) \
                - r_d * np.exp(-r_d * tau) * K * norm.cdf(d2 * cp)
            analytical_greeks['theta'] = analytical_formula * θ_shift
                
            # Gamma, Γ, is the change in an option's delta for a small change in the underlying assets price.
            # Gamma := ∂Δ/∂S ≈ (Δ(S_plus) − Δ(S_minus)) / (S_plus - S_minus)
            # In practice, gamma is  normalised to measure the change in Δ, for a 1% change in the underlying assets price.
            # Hence we have multiplied the analystical gamma formula by 'S * 0.01'
            analytical_formula = np.exp(-r_f * tau) * norm.pdf(d1)  / (S0 * σ * np.sqrt(tau)) # identical for calls and puts
            analytical_greeks['gamma'] = (0.01 * S0) * analytical_formula # normalised for 1% change
            
            # Rho, ρ, is the rate at which the price of an option changes relative to a change in the interest rate. 
            # In practice, Rho is normalised to measure the change in price for a 1% change in the underlying interest rate.
            analytical_formula = K * tau * np.exp(-r_f * tau) * norm.cdf(cp * d2)
            analytical_greeks['rho'] = analytical_formula * 0.01
            
            analytical_greeks = pd.DataFrame.from_dict(analytical_greeks)
            results['analytical_greeks'] = analytical_greeks

    
        if numerical_greeks_flag:
            numerical_greeks = {}
            
            if F != None:
                F_upshift = F*(1+Δ_shift)
                F_downshift = F*(1-Δ_shift)
            else:
                F_upshift = None
                F_downshift = None
            results_S0_plus = gk_price(S0=S0*(1+Δ_shift), tau=tau, r_f=r_f, r_d=r_d, cp=cp, K=K, σ=σ, F=F_upshift,analytical_greeks_flag=True) 
            X_S_plus, analytical_greeks_S0_plus = results_S0_plus['option_value'], results_S0_plus['analytical_greeks']
            results_S0_minus = gk_price(S0=S0*(1-Δ_shift), tau=tau, r_f=r_f, r_d=r_d, cp=cp, K=K, σ=σ, F=F_downshift,analytical_greeks_flag=True)
            X_S_minus, analytical_greeks_S0_minus = results_S0_minus['option_value'], results_S0_minus['analytical_greeks']
            
            numerical_greeks['spot_delta'] = (X_S_plus - X_S_minus) / (2 * S0 * Δ_shift)
            numerical_greeks['forward_delta'] = numerical_greeks['spot_delta'] / np.exp(-r_f * tau)
            
            X_σ_plus = gk_price(S0=S0, tau=tau, r_f=r_f, r_d=r_d, cp=cp, K=K, σ=σ+σ_shift, F=F)['option_value']
            X_σ_minus = gk_price(S0=S0, tau=tau, r_f=r_f, r_d=r_d, cp=cp, K=K, σ=σ-σ_shift, F=F)['option_value'] 
            numerical_greeks['vega'] = (X_σ_plus - X_σ_minus) / (2 * (σ_shift / 0.01))
            
            X_plus = gk_price(S0=S0, tau=tau+θ_shift, r_f=r_f, r_d=r_d, cp=cp, K=K, σ=σ, F=F)['option_value']     
            X_minus = gk_price(S0=S0, tau=tau-θ_shift, r_f=r_f, r_d=r_d, cp=cp, K=K, σ=σ, F=F)['option_value']  
            numerical_greeks['theta'] = (X_minus - X_plus) / 2 
            
            numerical_greeks['gamma'] = (analytical_greeks_S0_plus['spot_delta'] - analytical_greeks_S0_minus['spot_delta']) / (2 * (Δ_shift / 0.01))
            
            # This formulae will yield meaningfully different results on whether the forward rate is an input (or if it is calculated from the interest rate differential)
            X_ρ_plus = gk_price(S0=S0, tau=tau, r_f=r_f+ρ_shift, r_d=r_d, cp=cp, K=K, σ=σ+σ_shift, F=F)['option_value'] 
            X_ρ_minus = gk_price(S0=S0, tau=tau, r_f=r_f+ρ_shift, r_d=r_d, cp=cp, K=K, σ=σ-σ_shift, F=F)['option_value'] 
            numerical_greeks['rho'] = (X_ρ_plus - X_ρ_minus) / 2 
            
            numerical_greeks = pd.DataFrame.from_dict(numerical_greeks)
            results['numerical_greeks'] = numerical_greeks
            
        return results
            
def gk_solve_strike(S0: float, 
                   tau: float,                
                   r_f: float,                    
                   r_d: float,  
                   σ: float,                     
                   Δ: float,
                   Δ_convention: str,
                   F: float = None
                   ) -> float:
    """
    GarmanKohlhagen European FX option pricing formula.
    :param S0: FX spot, specified in # of units of domestic per 1 unit of foreign currency
    :param tau: time to np.expiry (in years)
    :param r_d: domestic risk free interest rate (annualised)
    :param r_f: foreign riskless interest rate (annualised)
    :param cp: option type (1 for call option (default), -1 for put option)
    :param σ: volatility
    :param F: market forward rate, if None is calculated under interest rate parity     
    :param Δ, signed delta, must be in range [0,0.5] for calls, [-0.5,0] for puts
    :param Δ_convention: delta quote convention {'regular_spot_Δ','regular_forward_Δ','premium_adjusted_spot_Δ','premium_adjusted_forward_Δ'}
    :return: strike, calculated from Δ
    """
  
    S0 = np.atleast_1d(S0).astype(float)
    tau = np.atleast_1d(tau).astype(float)
    r_f = np.atleast_1d(r_f).astype(float)
    r_d = np.atleast_1d(r_d).astype(float)
    σ = np.atleast_1d(σ).astype(float)
    Δ = np.atleast_1d(Δ).astype(float)
    
    # Sensical value checks
    assert (S0 > 0.0).all(), S0
    assert (tau > 0.0).all(), tau
    assert (σ > 0.0).all(), σ
    assert (Δ >= -0.5).all() and (Δ <= 0.5).all()
    assert Δ.shape == σ.shape
    cp = np.sign(Δ)
    Δ_convention = Δ_convention.replace('delta','Δ')
    assert Δ_convention in {'regular_spot_Δ',
                            'regular_forward_Δ',
                            'premium_adjusted_spot_Δ',
                            'premium_adjusted_forward_Δ'}, Δ_convention    

    if F is not None: 
        F = np.atleast_1d(F).astype(float)
        assert (F > 0.0).all() 

    result = np.zeros(shape=Δ.shape)
    if F == None:
        F = np.atleast_1d(S0 * np.exp((r_d - r_f) * tau)) # if not supplied, calculate the forward rate per interest rate parity 
    
    if Δ_convention in {'regular_spot_Δ','regular_forward_Δ'}:
        bool_cond = Δ == 0.5
        if bool_cond.any():
            # at-the-money Δ-neutral strike, for regular spot/forward Δ
            tmp = F * np.exp(0.5 * σ**2 * tau)
            result[bool_cond] = tmp[bool_cond]
        bool_cond = Δ != 0.5
        if bool_cond.any():        
            if Δ_convention == 'regular_spot_Δ':
                tmp = F * np.exp(-cp * norm.ppf(cp * Δ * np.exp(r_f * tau)) * σ * np.sqrt(tau) + (0.5 * σ**2) * tau)
                result[bool_cond] = tmp[bool_cond]
            elif Δ_convention == 'regular_forward_Δ':
                tmp = F * np.exp(-cp * norm.ppf(cp * Δ) * σ * np.sqrt(tau) + (0.5 * σ**2) * tau)
                result[bool_cond] = tmp[bool_cond]
    
    elif Δ_convention in {'premium_adjusted_spot_Δ','premium_adjusted_forward_Δ'}:
        bool_cond = Δ == 0.5
        if bool_cond.any():
            # at-the-money Δ-neutral strike, for premium adjusted spot/forward Δ
            tmp = F * np.exp(-1 * 0.5 * σ**2 * tau)
            result[bool_cond] = tmp[bool_cond]
            
        bool_cond = Δ != 0.5                
        if bool_cond.any(): 
            for i in range(len(Δ)):
                if Δ[i] != 0.5:
                    
                    # For premium adjusted quotes the solution must be solved numerically
                    # Please refer to 'A Guide to FX Options Quoting Conventions' by Uwe Wystub for full details
        
                    if Δ_convention == 'premium_adjusted_spot_Δ':
                        def solve_Δ(K, σ, cp, F, tau, Δ):
                            return np.exp(-r_f * tau) * (cp * K / F) * norm.cdf(cp * (np.log(F/K) - 0.5 * σ ** 2 * tau) / (σ * np.sqrt(tau))) - Δ                
                    elif Δ_convention == 'premium_adjusted_forward_Δ':
                        def solve_Δ(K, σ, cp, F, tau, Δ):
                            return (cp * K / F) * norm.cdf(cp * (np.log(F/K) - 0.5 * σ ** 2 * tau) / (σ * np.sqrt(tau))) - Δ
        
                    # Solve the upper bound, 'K_max' for the numerical solver
                    # The strike of a premium adjusted Δ-σ quote is ALWAYS below the regular (non premium adjusted) Δ-σ quote
                    # Hence, we analytically calculate the K, assuming the σ was a regular Δ quote
                    K_max = gk_solve_strike(S0=S0, 
                                            tau=tau,
                                            r_f=r_f, 
                                            r_d=r_d,  
                                            σ=σ[i], 
                                            Δ=Δ[i], 
                                            Δ_convention=Δ_convention.replace('premium_adjusted','regular'),
                                            F=F)
                    K_max = K_max.item()
    
                    # Put Option
                    if Δ[i] < 0: 
                        solution = root_scalar(solve_Δ, args=(σ[i], cp[i], F, tau, Δ[i]), x0=F, bracket=[0.00001, K_max])
                        
                    # Call Option
                    if Δ[i] > 0: 
                        # For the premimum adjusted call Δ, due to non-monotonicity, two strikes can be solved numerically for call options (but not put options).
                        # To avoid this, we solve a lower bound, 'K_min' to guarantee we get the correct solution in the numerical solver.
                        # The lower bound is the 'maximum' Δ, hence we numerically solve the maximum Δ
                        def solve_K_min(K, σ, cp, F, t):
                            d1 = (np.log(S0 / K) + (r_d - r_f + 0.5 * σ**2) * t) / (σ * np.sqrt(t))
                            d2 = d1 - σ * np.sqrt(t)
                            return σ * np.sqrt(t) * norm.cdf(d2) - norm.pdf(d2)
                        
                        try: 
                            solution = root_scalar(solve_K_min, args=(σ[i], cp[i], F, tau), x0=F ,bracket=[0.00001, K_max])
                        except ValueError('the numerical solve for K_min, for Δ', Δ[i], ', resulted in an error'):
                            pass
                    
                        if solution.converged:
                            K_min = solution.root
                        else:
                            raise ValueError('the numerical solver for K_min, for Δ', Δ[i], ', did not converge')
                        
                        try:
                            solution = root_scalar(solve_Δ, args=(σ[i], cp[i], F, tau, Δ[i]), x0=K_max, bracket=[K_min,K_max])
                        except ValueError('the numerical solve for premium adjusted strike for Δ', Δ[i], ', resulted in an error'):
                            print('This is likely due to an error in the input, for example typos or specifying the Δ_convention as spot Δ when it is actually forward Δ')
                        
                    if solution.converged:
                        result[i] = solution.root
                    else:
                        raise ValueError('the numerical solver for the premium adjusted strike for Δ', Δ[i], ', did not converge')
                
                
    return result
                    

def gk_solve_implied_σ(S0: float, 
                       tau: float, 
                       r_f: float,                       
                       r_d: float, 
                       cp: int,
                       K: float, 
                       X: float, 
                       σ_guess: float) -> float:
    """
    Solve the implied volatility using the Garman-Kohlhagen model for a European Vanilla FX option.

    Parameters:
    - S0 (float): Spot price
    - tau (float): Time to maturity in years
    - r_f (float): Foreign interest rate    
    - r_d (float): Domestic interest rate
    - cp (int): Option type (1 for call, -1 for put)
    - K (float): Strike price
    - X (float): Option price
    - σ_guess (float): Initial guess for volatility

    Returns:
    - float: Implied volatility

    Attempts to find implied volatility using Newton's method initially and falls back to Brent's method in case of failure.
    """
    try:
        # Try Netwon's method first (it's faster but less robust)
        return newton(lambda σ: (gk_price(S0=S0,tau=tau,r_f=r_f,r_d=r_d,cp=cp,K=K,σ=σ)['option_value']  - X), x0=σ_guess, tol=1e-4, maxiter=50)
    except RuntimeError:
        # Fallback to Brent's method
        try:
            return root_scalar(lambda σ: (gk_price(S0=S0,tau=tau,r_f=r_f,r_d=r_d,cp=cp,K=K,σ=σ)['option_value']  - X), bracket=[0.0001, 2], method='brentq').root
        except ValueError:
            return np.inf        
        
        
#%%

if __name__ == '__main__':
    pass

    S0=0.6438
    σ=0 #0.0953686
    r_d=0.05408
    r_f=0.04189
    tau=0.33403698
    cp=-1
    K=0.71000
    F = 0.646478
    p1 = gk_price(S0=S0,tau=tau,r_f=r_f,r_d=r_d,cp=cp,K=K,σ=σ,F=F)['option_value']
    print('F specified:', p1)
    p2 = gk_price(S0=S0,tau=tau,r_f=r_f,r_d=r_d,cp=cp,K=K,σ=σ)['option_value']
    print('IR parity', p2)


    # check if instrinisc value is returned. Use IR parity. 
    p3 = gk_price(S0=S0,tau=tau,r_f=r_f,r_d=r_d,cp=cp,K=K,σ=σ)['option_value']
    print('if tau=0 intrinisc?:', p3)    















# -*- coding: utf-8 -*-


import os
import pathlib

if __name__ == "__main__":
    os.chdir(pathlib.Path(__file__).parent.parent.parent.resolve()) # path to ./frm/
    print(__file__.split('\\')[-1], os.getcwd())
    
from pricing_engine.garman_kohlhagen import garman_kohlhagen
import numpy as np
from scipy.stats import norm
from scipy.integrate import quad
from scipy.optimize import minimize, newton, root_scalar
from joblib import Parallel, delayed
import time
from numba import jit

def simHeston(n, s0, v0, mu, kappa, theta, sigma, rho, delta, no=None, method=0):
    
    
    # https://www.degruyter.com/document/doi/10.1515/math-2017-0058/html?lang=en
    # E + M scheme or QE scheme appear to be the quickest
    
    if no is None:
        no = np.random.normal(0, 1, (int(n/delta), 2))

    if len(no[:, 0]) != int(n/delta):
        raise ValueError('Size of no is inappropriate. Length of no[:, 0] should be equal to n/delta.')

    t = np.arange(0, np.ceil(n/delta)+1)
    sizet = len(t)
    x = np.zeros((sizet, 2))
    C = np.array([1, rho, rho, 1]).reshape(2, 2)
    u = normalcorr(C, no) * np.sqrt(delta)

    if method == 0:
        x[0, :] = [np.log(s0), v0]
        phiC = 1.5

        for i in range(1, sizet):
            m = theta + (x[i-1, 1] - theta) * np.exp(-kappa * delta)
            s2 = (x[i-1, 1] * sigma ** 2 * np.exp(-kappa * delta) / kappa * (1 - np.exp(-kappa * delta))
                  + theta * sigma ** 2 / (2 * kappa) * (1 - np.exp(-kappa * delta)) ** 2)
            phi = s2 / m ** 2
            gamma1, gamma2 = 0.5, 0.5
            K0 = -rho * kappa * theta / sigma * delta
            K1 = gamma1 * delta * (kappa * rho / sigma - 0.5) - rho / sigma
            K2 = gamma2 * delta * (kappa * rho / sigma - 0.5) + rho / sigma
            K3 = gamma1 * delta * (1 - rho ** 2)
            K4 = gamma2 * delta * (1 - rho ** 2)

            if phi <= phiC:
                b2 = 2 / phi - 1 + np.sqrt(2 / phi) * np.sqrt(2 / phi - 1)
                a = m / (1 + b2)
                x[i, 1] = a * (np.sqrt(b2) + no[i-1, 1]) ** 2
            else:
                p = (phi - 1) / (phi + 1)
                beta = (1 - p) / m

                if 0 <= norm.cdf(no[i-1, 1], 0, 1) <= p:
                    x[i, 1] = 0
                elif p < norm.cdf(no[i-1, 1], 0, 1) <= 1:
                    x[i, 1] = 1 / beta * np.log((1 - p) / (1 - norm.cdf(no[i-1, 1], 0, 1)))

            x[i, 0] = (x[i-1, 0] + mu * delta + K0 + K1 * x[i-1, 1] + K2 * x[i, 1]
                       + np.sqrt(K3 * x[i-1, 1] + K4 * x[i, 1]) * no[i-1, 0])

        x[:, 0] = np.exp(x[:, 0])
    else:
        x[0, :] = [s0, v0]

        for i in range(1, sizet):
            if method == 1:
                x[i-1, 1] = max(x[i-1, 1], 0)
            else:
                x[i-1, 1] = max(x[i-1, 1], -x[i-1, 1])

            x[i, 1] = (x[i-1, 1] + kappa * (theta - x[i-1, 1]) * delta
                       + sigma * np.sqrt(x[i-1, 1]) * u[i-1, 1])
            x[i, 0] = x[i-1, 0] + x[i-1, 0] * (mu * delta + np.sqrt(x[i-1, 1]) * u[i-1, 0])

    return x


def normalcorr(C, no):
    M = np.linalg.cholesky(C).T
    return (M @ no.T).T


def simGBM(n, x0, mu, sigma, delta, no=None, method=0):
    if no is None:
        no = np.random.normal(0, 1, int(np.ceil(n/delta)))
    else:
        no = no.flatten()
        if len(no) != int(np.ceil(n/delta)):
            raise ValueError('Error: length(no) <> n/delta')

    if method == 1: # Euler scheme
        x = x0 * np.cumprod(1 + mu * delta + sigma * delta**0.5 * no)
    elif method == 2: # Milstein scheme
        x = x0 * np.cumprod(1 + mu * delta + sigma * delta**0.5 * no +
                           0.5 * sigma**2 * delta * (no**2 - 1))
    elif method == 3: # 2nd order Milstein scheme
        x = x0 * np.cumprod(1 + mu * delta + sigma * delta**0.5 * no +
                           0.5 * sigma**2 * delta * (no**2 - 1) +
                           mu * sigma * no * (delta**1.5) + 0.5 * (mu**2) * (delta**2))
    else: # Direct integration
        x = x0 * np.exp(np.cumsum((mu - 0.5 * sigma**2) * delta + sigma * delta**0.5 * no))

    # Add starting value
    x = np.insert(x, 0, x0)
    return x


def pdfHestonInt(x, theta, kappa, sigma, rho, t, px):
    Gamma = kappa + complex(0,1) * rho * sigma * px
    Gamma2 = Gamma * Gamma
    Omega = np.sqrt(Gamma2 + sigma ** 2 * (px ** 2 - complex(0,1) * px))
    Omega2 = Omega * Omega
    Ft1 = kappa * theta / (sigma ** 2) * Gamma * t

    ncoeff = Omega2 - Gamma2 + 2 * kappa * Gamma
    coeff = ncoeff / (2 * kappa * Omega)
    Ot = Omega * t / 2
    Err = np.cosh(Ot)

    Ft2 = np.log(Err + coeff * np.sinh(Ot))
    Ft = Ft1 - 2 * kappa * theta / (sigma ** 2) * Ft2
    Eksp = np.exp(Ft + complex(0,1) * x * px)
    return np.real(Eksp)

def pdfHeston(x, theta, kappa, sigma, rho, t, par=0):
    n = len(x)
    y = np.zeros(n)
    pxm = 100

    for i in range(n):
        y[i] = 1 / (2 * np.pi) * quad(lambda s: pdfHestonInt(x[i], theta, kappa, sigma, rho, t, s), -100, 100, epsrel=1e-8)[0]

    if par == 1:
        import matplotlib.pyplot as plt
        plt.plot(x, y)
        plt.show()

    return y

@jit(nopython=True)
def HestonVanillaInt(phi, m, s, k, v0, vv, rd, rf, tau, kappa, theta, lambda_, rho):
    a = kappa * theta
    u = [0.5, -0.5]
    b = [kappa + lambda_ - rho * vv, kappa + lambda_]
    x = np.log(s)
    y = np.log(k)

    d = np.sqrt((complex(0,1) * rho * vv * phi - b[m-1]) ** 2 - vv ** 2 * (2 * u[m-1] * phi * complex(0,1) - phi ** 2))
    g = (b[m-1] - rho * vv * phi * complex(0,1) - d) / (b[m-1] - rho * vv * phi * complex(0,1) + d)
    D = (b[m-1] - rho * vv * phi * complex(0,1) - d) / (vv ** 2) * ((1 - np.exp(-d * tau)) / (1 - g * np.exp(-d * tau)))
    C = (rd - rf) * phi * complex(0,1) * tau + a / (vv ** 2) * ((b[m-1] - rho * vv * phi * complex(0,1) - d) * tau - 2 * np.log((1 - g * np.exp(-d * tau)) / (1 - g)))
    f = np.exp(C + D * v0 + complex(0,1) * phi * x)
    F = np.real(np.exp(-complex(0,1) * phi * y) * f / (complex(0,1) * phi))

    return F

def HestonVanilla(cp, s, k, v0, vv, rd, rf, tau, kappa, theta, lambda_, rho):

    P1 = 0.5 + 1/np.pi * quad(lambda phi: HestonVanillaInt(phi, 1, s, k, v0, vv, rd, rf, tau, kappa, theta, lambda_, rho), 0, np.inf, epsrel=1e-8)[0]
    P2 = 0.5 + 1/np.pi * quad(lambda phi: HestonVanillaInt(phi, 2, s, k, v0, vv, rd, rf, tau, kappa, theta, lambda_, rho), 0, np.inf, epsrel=1e-8)[0]

    Pplus = (1 - cp) / 2 + cp * P1
    Pminus = (1 - cp) / 2 + cp * P2

    P = cp * (s * np.exp(-rf * tau) * Pplus - k * np.exp(-rd * tau) * Pminus)
    return P


def HestonVanillaFitSmile(delta, marketvols, spot, rd, rf, tau, cp):
    
    strikes = garman_kohlhagen(S=spot,σ=marketvols,r_f=rf,r_d=rd,tau=tau,cp=cp,Δ=delta,task='strike',Δ_convention='regular_spot_Δ')

    nostrikes = len(strikes)
    v0 = (marketvols[nostrikes // 2])**2
    kappa = 1.5
    initparam = [2 * np.sqrt(v0), 2 * v0, 0]

    def HestonVanillaSSE(param, cp, spot, strikes, v0, marketvols, rd, rf, tau, kappa):        
        vv, theta, rho = param
        if vv < 0 or theta < 0 or abs(rho) > 1:
            return np.inf
        P = np.zeros(nostrikes)
        IV = np.zeros(nostrikes)
        for i in range(nostrikes):
            P[i] = HestonVanilla(cp, spot, strikes[i], v0, vv, rd, rf, tau, kappa, theta, 0, rho)
            IV[i] = ImplVolFX(P[i], spot, strikes[i], rd, rf, tau, cp, marketvols[i])
        return np.sum((marketvols - IV)**2)

    def ImplVolFX(X, S, K, rd, rf, tau, cp, σ_guess):
        try:
            # Try Netwon's method first
            return newton(lambda σ: (garman_kohlhagen(S=spot,σ=σ,r_f=rf,r_d=rd,tau=tau,cp=cp,K=K,Δ=delta,task='px') - X), x0=σ_guess, tol=1e-4, maxiter=50)
        except RuntimeError:
            # Fallback to Brent's method
            return root_scalar(lambda σ: (garman_kohlhagen(S=spot,σ=σ,r_f=rf,r_d=rd,tau=tau,cp=cp,K=K,Δ=delta,task='px') - X), bracket=[0.001, 2], method='brentq').root

    res = minimize(lambda param: HestonVanillaSSE(param, cp, spot, strikes, v0, marketvols, rd, rf, tau, kappa), initparam)
    vv, theta, rho = res.x
    P = np.zeros(nostrikes)
    IV = np.zeros(nostrikes)
    
    for i in range(nostrikes):
        P[i] = HestonVanilla(cp, spot, strikes[i], v0, vv, rd, rf, tau, kappa, theta, 0, rho)
        IV[i] = ImplVolFX(P[i], spot, strikes[i], rd, rf, tau, cp, marketvols[i])
 
    SSE = HestonVanillaSSE(res.x, cp, spot, strikes, v0, marketvols, rd, rf, tau, kappa)
    return v0, vv, kappa, theta, rho, IV, SSE







# -*- coding: utf-8 -*-


if __name__ == "__main__":
    import os
    import pathlib
    os.chdir(pathlib.Path(__file__).parent.parent.parent.parent.resolve())     
    print('__main__ - current working directory:', os.getcwd())
    
import numpy as np
from numba import jit, prange
import timeit
import time

from frm.frm.pricing_engine.monte_carlo_generic import generate_rand_nbs


def simulate_gbm_path(initial_px: np.array,
                      drift: np.array, 
                      forward_volatility: np.array,
                      timestep_length: np.array, 
                      rand_nbs: np.array):
    """
    Simulates the path of a Geometric Brownian Motion (GBM) over a defined period set.

    This function simulates the path of a stochastic process following GBM, given initial prices,
    drift rates, volatilities, and random numbers for each period. It supports vectorized operations
    allowing simulation of multiple paths in parallel.

    Parameters:
    initial_px (np.array): An array of initial prices for each simulation path.
    drift (np.array): An array of drift rates for each time step.
    forward_volatility (np.array): An array of volatilities for each time step.
    timestep_length (np.array): An array representing the length of each step
    rand_nbs (np.array): A 3D array of random numbers; shape=(number of time steps, number of random variables, number of simulations)

    Returns:
    np.array: A 2D array where each row represents a simulated GBM path and each column
              represents a time step in the path.

    Raises:
    AssertionError: If the shapes of period_length, period_drift, period_forward_volatility,
                    and rand_nbs do not align as expected.

    Notes:
    - The function assumes that period_length, period_drift, and period_forward_volatility have
      the same shape and that the number of periods matches the number of columns in rand_nbs.
    - It simulates the path by sequentially updating the price for each period using the GBM formula.
    """    
    
    assert timestep_length.ndim == 1
    
    # Reshape inputs to 2D arrays so to support parallel simulation of multiple underlyings 
    drift = np.atleast_1d(drift)
    forward_volatility = np.atleast_1d(forward_volatility)
    initial_px = np.atleast_1d(initial_px)
    
    if drift.ndim == 1:
        drift = drift.reshape(-1, 1)
    if forward_volatility.ndim == 1:
        forward_volatility = forward_volatility.reshape(-1, 1)  
    if initial_px.ndim == 1:
        initial_px = initial_px.reshape(1, -1)          
    
    # Shape checks
    nb_timesteps = timestep_length.shape[0]
    assert nb_timesteps == drift.shape[0]
    assert nb_timesteps == forward_volatility.shape[0]
    assert nb_timesteps == rand_nbs.shape[0]
    assert initial_px.shape[0] == 1
    
    nb_rand_vars = initial_px.shape[1]
    assert nb_rand_vars == drift.shape[1]
    assert nb_rand_vars == forward_volatility.shape[1]
    assert nb_rand_vars == rand_nbs.shape[1]
    
    # Initialise shape and set initial value of underlying(s)
    x = np.zeros((rand_nbs.shape[0]+1, rand_nbs.shape[1], rand_nbs.shape[2]))
    x[0,:,:] = initial_px
    
    # Simulate per geometric brownian motion process
    for i in range(0, nb_timesteps):
        x[i+1,:,:] = x[i,:,:] * np.exp((drift[i,:] - 0.5 * (forward_volatility[i,:] ** 2)) * timestep_length[i] + forward_volatility[i,:] * np.sqrt(timestep_length[i]) * rand_nbs[i,:,:])
        
    return x


@jit(nopython=True)
def simulate_gbm_path_numba(initial_px: np.array,
                            drift: np.array, 
                            forward_volatility: np.array,
                            timestep_length: np.array, 
                            rand_nbs: np.array):
    
    assert timestep_length.shape == drift.shape
    assert timestep_length.shape == forward_volatility.shape
    assert timestep_length.shape[0] == rand_nbs.shape[0]
    assert initial_px.shape == drift.shape
    
    nb_periods = len(timestep_length)
    
    x = np.zeros((rand_nbs.shape[0],nb_periods+1))
    x[:,0,] = initial_px
    
    for i in range(0, nb_periods):
        x[:,i+1] = x[:,i] * np.exp((drift[i] - 0.5 * (forward_volatility[i] ** 2)) * timestep_length[i] + forward_volatility[i] * np.sqrt(timestep_length[i]) * rand_nbs[:,i])
        
    return x
            

            
if __name__ == "__main__":
            

    initial_px = np.array([0.6629])
    drift = np.array([0.95*0.01, 0.55*0.01, -0.02*0.01, -0.34*0.01])
    forward_volatility = np.array([9.73*0.01, 10.10*0.01, 9.75*0.01, 10.97*0.01])
    timestep_length = np.array([0.512328767, 0.498630137, 0.501369863, 0.498630137])
    nb_periods=len(timestep_length)
    
    rand_nbs = generate_rand_nbs(nb_periods=nb_periods,
                                 nb_rand_vars=1,
                                 nb_simulations=1000 * 1000)
    
    
    # results = simulate_gbm_path_numba(initial_px=initial_px,
    #                                   drift=drift,
    #                                   forward_volatility=forward_volatility,
    #                                   timestep_length=timestep_length,
    #                                   rand_nbs=rand_nbs)
    
        
    t1 = time.time()
    results = simulate_gbm_path(initial_px=initial_px,
                                      drift=drift,
                                      forward_volatility=forward_volatility,
                                      timestep_length=timestep_length,
                                      rand_nbs=rand_nbs)
    
    # t2 = time.time()
    # results = simulate_gbm_path_numba(initial_px=initial_px,
    #                                   drift=drift,
    #                                   forward_volatility=forward_volatility,
    #                                   timestep_length=timestep_length,
    #                                   rand_nbs=rand_nbs)
    
    t3 = time.time()
    
    
    print("----------")
    print("GBM:", t2-t1)
    print("GBM numba:", t3-t2)
    #print("GBM numba_prange:", t4-t3)    
    
    
    
    
    # # Time the original function
    # time_original = timeit.timeit(
    #     """simulate_gbm_path(initial_px=initial_px,
    #                                       period_drift=period_drift,
    #                                       period_forward_volatility=period_forward_volatility,
    #                                       period_length=period_length,
    #                                       rand_nbs=rand_nbs)""",
    #     globals=globals(),
    #     number=10
    # )
    
    # # Time the Numba-optimized function (first call compiles the function)
    # results = simulate_gbm_path_numba(initial_px=initial_px,
    #                                   period_drift=period_drift,
    #                                   period_forward_volatility=period_forward_volatility,
    #                                   period_length=period_length,
    #                                   rand_nbs=rand_nbs)
    
    # time_numba = timeit.timeit(
    #     """simulate_gbm_path_numba(initial_px=initial_px,
    #                                       period_drift=period_drift,
    #                                       period_forward_volatility=period_forward_volatility,
    #                                       period_length=period_length,
    #                                       rand_nbs=rand_nbs)""",
    #     globals=globals(),
    #     number=10
    # )
    
    #print(f"Time taken by original function: {time_original:.6f} seconds")
    #print(f"Time taken by Numba-optimized function: {time_numba:.6f} seconds")

#%%


            
            
            
            

def simulate_gbm(x0, mu, σ, T, dt, rand_nbs=None, method=0):
    """
    Simulate Geometric Brownian Motion (GBM).
    
    Parameters:
    x0 (float): Initial value of the process.
    mu (float): Drift.
    σ (float): Volatility.
    T (float): Time end point
    dt (float): Time step size.
    no (np.ndarray, optional): Array of normally distributed pseudorand nbs. Defaults to None.
    method (int, optional): Method for simulation. Defaults to 0.
        0: Direct integration
        1: Euler scheme
        2: Milstein scheme 1st order
        3: Milstein scheme 2nd order

    Returns:
    np.ndarray: Vector containing the simulated trajectory of the GBM.
    
    References
    [1]: Rouah, F.D. (2013). Euler Scheme for the Black-Scholes Model.
    """
    
    if rand_nbs is None:
        rand_nbs = np.rand.normal(0, 1, int(np.ceil(T / dt)))
    
    rand_nbs = np.asarray(rand_nbs).flatten()
    
    if len(rand_nbs) != int(np.ceil(T / dt)):
        raise ValueError("Error: length(no) != n/dt")
    
    if method == 0:
        x = x0 * np.exp(np.cumsum((mu - 0.5 * σ ** 2) * dt + σ * np.sqrt(dt) * rand_nbs))    
    elif method == 1:
        x = x0 * np.cumprod(1 + mu * dt + σ * np.sqrt(dt) * rand_nbs)
    elif method == 2:
        x = x0 * np.cumprod(1 + mu * dt + σ * np.sqrt(dt) * rand_nbs +
                            0.5 * σ ** 2 * dt * (rand_nbs ** 2 - 1))
    elif method == 3:
        x = x0 * np.cumprod(1 + mu * dt + σ * np.sqrt(dt) * rand_nbs +
                            0.5 * σ ** 2 * dt * (rand_nbs ** 2 - 1) +
                            mu * σ * rand_nbs * (dt ** 1.5) + 0.5 * (mu ** 2) * (dt ** 2))

    return np.concatenate(([x0], x))


@jit(nopython=True, parallel=True)
def simulate_gbm_numba_prange(x0, mu, σ, T, dt, rand_nbs=None, method=0):
    steps = int(np.ceil(T / dt))
    
    if rand_nbs is None:
        rand_nbs = np.rand.normal(0, 1, steps)
    
    rand_nbs = np.asarray(rand_nbs).flatten()
    
    if len(rand_nbs) != steps:
        raise ValueError("Error: length(no) != n/dt")
    
    sqrt_dt = np.sqrt(dt)
    
    x = np.empty(steps)
    
    if method == 0:
        for i in prange(steps):
            x[i] = np.exp((mu - 0.5 * σ ** 2) * dt + σ * sqrt_dt * rand_nbs[i])
        x = np.cumprod(x)
    elif method == 1:
        for i in prange(steps):
            x[i] = 1 + mu * dt + σ * sqrt_dt * rand_nbs[i]
        x = np.cumprod(x)
    elif method == 2:
        for i in prange(steps):
            x[i] = 1 + mu * dt + σ * sqrt_dt * rand_nbs[i] + 0.5 * σ ** 2 * (dt * (rand_nbs[i] ** 2 - 1))
        x = np.cumprod(x)
    elif method == 3:
        for i in prange(steps):
            x[i] = 1 + mu * dt + σ * sqrt_dt * rand_nbs[i] + 0.5 * σ ** 2 * (dt * (rand_nbs[i] ** 2 - 1)) + \
                   mu * σ * rand_nbs[i] * (dt ** 1.5) + 0.5 * (mu ** 2) * (dt ** 2)
        x = np.cumprod(x)
    
    result = np.empty(steps + 1)
    result[0] = x0
    for i in prange(steps):
        result[i + 1] = x0 * x[i]
    
    return result



@jit(nopython=True)
def simulate_gbm_numba(x0, mu, sigma, T, dt, rand_nbs=None, method=0):
    
    steps = int(np.ceil(T / dt))
    
    if rand_nbs is None:
        rand_nbs = np.rand.normal(0, 1, steps)
    
    if len(rand_nbs) != steps:
        raise ValueError("Error: length(no) != steps/dt")
    
    x = np.empty(steps+1)
    x[0] = x0
    
    if method == 0:
        for i in range(1, steps+1):
            x[i] = x[i-1] * np.exp((mu - 0.5 * sigma ** 2) * dt + sigma * np.sqrt(dt) * rand_nbs[i-1])
    elif method == 1:  # euler_scheme
        for i in range(1, steps+1):
            x[i] = x[i-1] * (1 + mu * dt + sigma * np.sqrt(dt) * rand_nbs[i-1])
    elif method == 2:  # milstein_scheme_1st_order
        for i in range(1, steps+1):
            x[i] = x[i-1] * (1 + mu * dt + sigma * np.sqrt(dt) * rand_nbs[i-1] +
                            0.5 * sigma ** 2 * dt * (rand_nbs[i-1] ** 2 - 1))
    elif method == 3:  # milstein_scheme_2nd_order
        for i in range(1, steps+1):
            x[i] = x[i-1] * (1 + mu * dt + sigma * np.sqrt(dt) * rand_nbs[i-1] +
                            0.5 * sigma ** 2 * dt * (rand_nbs[i-1] ** 2 - 1) +
                            mu * sigma * rand_nbs[i-1] * (dt ** 1.5) + 0.5 * (mu ** 2) * (dt ** 2))    
    return x


#%%



if __name__ == "__main__":
    
    # Numba is much faster after it has been compiled.
    # Key is having a problem where it gets compiled

    import timeit
    import time

    # Parameters
    x0 = 100
    mu = 0.05
    σ = 0.2
    n = 10000
    dt = 0.01
    rand_nbs = np.rand.normal(0, 1, int(np.ceil(n / dt)))
    
    
    t1 = time.time()
    x = simulate_gbm(x0, mu, σ, n, dt, rand_nbs=rand_nbs, method=0)
    t2 = time.time()
    y = simulate_gbm_numba(x0, mu, σ, n, dt, rand_nbs=rand_nbs, method=0)
    t3 = time.time()
    y = simulate_gbm_numba_prange(x0, mu, σ, n, dt, rand_nbs=rand_nbs, method=0)
    t4 = time.time()
    
    print("----------")
    print("GBM:", t2-t1)
    print("GBM numba:", t3-t2)
    print("GBM numba_prange:", t4-t3)    




    # Time the original function
    time_original = timeit.timeit(
        'simulate_gbm(x0, mu, σ, n, dt, rand_nbs=rand_nbs, method=0)',
        globals=globals(),
        number=100
    )
    
    # Time the Numba-optimized function (first call compiles the function)
    simulate_gbm_numba(x0, mu, σ, n, dt, rand_nbs=rand_nbs, method=0)
    
    time_numba = timeit.timeit(
        'simulate_gbm_numba(x0, mu, σ, n, dt, rand_nbs=rand_nbs, method=0)',
        globals=globals(),
        number=100
    )

    # Time the Numba-optimized function (first call compiles the function)
    simulate_gbm_numba_prange(x0, mu, σ, n, dt, rand_nbs=rand_nbs, method=0)
        
    time_numba_prange = timeit.timeit(
        'simulate_gbm_numba_prange(x0, mu, σ, n, dt, rand_nbs=rand_nbs, method=0)',
        globals=globals(),
        number=100
    )    
    
    # Display results
    print(f"Time taken by original function: {time_original:.6f} seconds")
    print(f"Time taken by Numba-optimized function: {time_numba:.6f} seconds")
    print(f"Time taken by Numba-Prange-optimized function: {time_numba_prange:.6f} seconds")















# -*- coding: utf-8 -*-
import numpy as np
MAX_SIMULATIONS_PER_LOOP = 100e6

def generate_rand_nbs(nb_steps: int,
                      nb_rand_vars: int=1,
                      nb_simulations: int=None,  
                      flag_apply_antithetic_variates: bool=None,
                      random_seed=0):
    
    """
    Generate random numbers for Monte Carlo simulations with an option to apply antithetic variates.

    This function generates normally distributed random numbers for use in Monte Carlo simulations.
    It can optionally generate antithetic variates to help in variance reduction. The function is
    designed to support simulations across multiple periods and can handle a large number of simulations.

    Parameters:
    nb_steps (int): The number of periods for which random numbers need to be generated.
    nb_rand_vars (int) The number of random numbers
    nb_simulations (int, optional): The total number of simulations. Default is 100,000.
    flag_apply_antithetic_variates (bool, optional): Flag to indicate whether antithetic variates should
                                                     be applied for variance reduction. Default is True.

    Returns:
    np.array: A 2D array of random numbers. Each row corresponds to a simulation path, and each
              column corresponds to a period.

    Raises:
    ValueError: If the number of simulations exceeds 10 million, as this may lead to memory issues.

    Notes:
    - If `flag_apply_antithetic_variates` is True, the function generates half the number of 
      normal simulations and then creates antithetic variates for these. The results are concatenated
      to form the final set of simulations.
    - The function ensures that the total number of simulations (normal and antithetic) matches the
      specified `nb_simulations`.
    - The generated random numbers follow a standard normal distribution (mean 0, standard deviation 1).
    """
    
    np.random.seed(random_seed)
    
    if nb_simulations is None:
        nb_simulations = 100 * 1000 
        
    if flag_apply_antithetic_variates is None:
        flag_apply_antithetic_variates = False
    
    assert isinstance(nb_steps, int), type(nb_steps) 
    assert isinstance(nb_rand_vars, int), type(nb_rand_vars)
    assert isinstance(nb_simulations, int), type(nb_simulations)
    assert isinstance(flag_apply_antithetic_variates, bool)
    assert nb_steps >= 1, nb_steps
    assert nb_rand_vars >= 1, nb_rand_vars
    assert nb_simulations >= 1, nb_simulations
    
    if (nb_steps * nb_simulations) > MAX_SIMULATIONS_PER_LOOP:
        raise ValueError("Too many steps & simulations for one refresh; may lead to memory leak")    
        
    if flag_apply_antithetic_variates and nb_simulations == 1:
        raise ValueError("Antithetic variates requiries >=2 simulations") 
        
    if flag_apply_antithetic_variates:
        nb_antithetic_variate_simulations = nb_simulations // 2
        nb_normal_simulations = nb_simulations - nb_antithetic_variate_simulations
        rand_nbs_normal = np.random.normal(0, 1, (nb_steps, nb_rand_vars, nb_normal_simulations)) # standard normal random numbers
        rand_nbs_antithetic_variate = -1 * rand_nbs_normal[:,:,:nb_antithetic_variate_simulations]
        rand_nbs = np.concatenate([rand_nbs_normal, rand_nbs_antithetic_variate], axis=2)
    else:    
        rand_nbs = np.random.normal(0, 1, (nb_steps, nb_rand_vars, nb_simulations))
    
    return rand_nbs


if __name__ == "__main__":
    rand_nbs = generate_rand_nbs(nb_steps=20,
                                 nb_rand_vars=1,
                                 nb_simulations=1000,
                                 flag_apply_antithetic_variates=True)





# -*- coding: utf-8 -*-

import pandas as pd
import numpy as np
from frm.frm.schedule.business_day_calendar import get_calendar
from frm.frm.schedule.tenor import calc_tenor_date, get_spot_offset
from frm.frm.schedule.daycounter import DayCounter

def convert_column_to_consistent_data_type(df: pd.DataFrame):
    for col in df.columns:
        if df[col].apply(isinstance, args=(float,)).all():
            df[col] = pd.to_numeric(df[col])
            
    return df


def copy_errors_and_warnings_to_input(df_processed, df_input):
    
    for v in ['errors','warnings']:
    
        internal_ids = df_processed.loc[df_processed[v] != '','internal_id'].to_list()
        for i in internal_ids:
            mask_processed = df_processed['internal_id'] == i
            mask_input = df_input['internal_id'] == i
            
            df_input.loc[mask_input,v] = df_processed.loc[mask_processed,v].iloc[0]
        
    return df_input


def clean_input_dataframe(df):    
    
    if 'internal_id' not in df.columns:
        df = df.reset_index().rename(columns={'index': 'internal_id'})
        
    df = df.dropna(subset=[col for col in df.columns if col != 'internal_id'], how='all')
    df.reset_index(drop=True, inplace=True)
    
    df = df.rename(columns=lambda x: x.strip() if isinstance(x, str) else x)
    df = df.applymap(lambda x: x.lower() if isinstance(x, str) else x)

    if 'errors' not in df.columns:
        df.insert(0, 'errors', value='')
    if 'warnings' not in df.columns:
        df.insert(1, 'warnings', value='')
    
    return df


def move_col_after(df, col_to_move, ref_col):
    cols = df.columns.tolist()
    cols.insert(cols.index(ref_col) + 1, cols.pop(cols.index(col_to_move)))
    return df[cols]




def generic_market_data_input_cleanup_and_validation(df : pd.DataFrame,
                                                     spot_offset: bool=True):
    df_input = df.copy()

    # mandatory column validation  
    mandatory_columns = [
        'curve_date',
        'curve_ccy',
    ]
    df = df.dropna(axis=0, subset=mandatory_columns) # drop rows with blanks in mandatory columns
    missing_mandatory_columns = [col for col in mandatory_columns if col not in df.columns.to_list()]
    if len(missing_mandatory_columns) > 0:
        df['errors'] += f'missing mandatory columns: {missing_mandatory_columns}\n'
        return df

    # tenor input validation 
    if 'tenor_date' not in df.columns and 'tenor_name' not in df.columns:
        df['errors'] += f'a tenor input via tenor_name or tenor_date is mandatory\n'
    elif 'tenor_date' not in df.columns and 'tenor_name' in df.columns:
        df['tenor_date'] = np.nan
        df = move_col_after(df=df, col_to_move='tenor_date', ref_col='tenor_name')

    df['calendar'] = np.nan
    df['base_ccy'] = np.nan
    df['quote_ccy'] = np.nan
    df = move_col_after(df=df, col_to_move='base_ccy', ref_col='curve_ccy')
    df = move_col_after(df=df, col_to_move='quote_ccy', ref_col='curve_ccy')    
    
    # create a dictionary of all holiday calendars required
    curve_ccy_cal_dict = {}
    for curve_ccy in df['curve_ccy'].dropna().unique():
        if len(curve_ccy) == 3:
            curve_ccy_cal_dict[curve_ccy] = get_calendar(ccys=curve_ccy)
        elif len(curve_ccy) == 6:
            curve_ccy_cal_dict[curve_ccy] = get_calendar(ccys=[curve_ccy[:3],curve_ccy[3:]])    

    # row level validation
    for i,row in df.iterrows():
        
        field = 'curve_date'
        if not isinstance(pd.Timestamp(row[field]), pd.Timestamp):
            df.at[i,'errors'] += field + ' value is not a valid input\n'
            
        if pd.isna(row['tenor_name']) and pd.isna(row['tenor_date']):
            df.at[i,'errors'] += 'a tenor input in tenor_name or tenor_date is mandatory\n'
    
        if pd.isna(row['tenor_name']) and not pd.isna(row['tenor_date']):
            if not isinstance(pd.Timestamp(row['tenor_date']), pd.Timestamp):
                df.at[i,'errors'] += 'tenor_date' + ' value is not a valid input\n'

        field = 'curve_ccy'
        if not isinstance(row[field], str):
            df.at[i,'errors'] += field + ' value is not a valid input\n'
        else:
            if len(row[field]) not in {3,6}:
                df.at[i,'errors'] += field + ' value is not a valid input\n'
            else:
                if len(row[field]) == 3:
                    df.at[i,'calendar'] = curve_ccy_cal_dict[row[field]]
                elif len(row[field]) == 6:
                    df.at[i,'calendar'] = curve_ccy_cal_dict[row[field]]
                    df.at[i,'base_ccy'] = row[field][:3]
                    df.at[i,'quote_ccy'] = row[field][3:]  

    for i,row in df.iterrows():        
        if pd.isna(row['tenor_date']) and pd.notna(row['tenor_name']): 
            tenor_date, tenor_name_cleaned, spot_date = calc_tenor_date(row['curve_date'], row['tenor_name'], row['curve_ccy'], row['calendar'], spot_offset)
            df.at[i,'tenor_date'] = tenor_date
            df.at[i,'tenor_name'] = tenor_name_cleaned
        
    df = df.drop(['calendar'], axis=1)  
    
    if 'day_count_basis' not in df.columns:
        day_counter = DayCounter()
        df['day_count_basis'] = day_counter.day_count_basis
        df['tenor_years'] = day_counter.year_fraction(df['curve_date'], df['tenor_date'])
    else:
        df['tenor_years'] = np.nan
        for i,row in df.iterrows():
            day_counter = DayCounter(day_count_basis=row['day_count_basis'])
            df.at[i,'day_count_basis'] = day_counter.day_count_basis
            df.at[i,'tenor_years'] = day_counter.year_fraction(df.at[i,'curve_date'], df.at[i,'tenor_date'])
        
    df = move_col_after(df, 'day_count_basis', 'tenor_date')
    df = move_col_after(df, 'tenor_years', 'day_count_basis')
    
    # If column values are a consistent type, set the dataframe column type to that
    for col in df.columns:
        if df[col].apply(isinstance, args=(float,)).all():
            df[col] = pd.to_numeric(df[col])
    
    # In code we use Δ in all instances, not the word 'delta'
    df = df.applymap(lambda x: x.replace('delta', 'Δ') if isinstance(x, str) else x)
    df.columns = [col.replace('delta', 'Δ') for col in df.columns]             
                       
    return df




    


#%%
"""
Created on Thu Nov 30 2018
Merton Model and convergence obtained with the COS method
@author: Lech A. Grzelak
"""
import numpy as np
import matplotlib.pyplot as plt
import scipy.stats as st
import enum
import scipy.optimize as optimize

# Set i= imaginary number
i = 1j


def CallPutOptionPriceCOSMthd(cf,cp,S0,r,tau,K,N,L):
    # cf - Characteristic function as a functon, in the book denoted by phi
    # cp - C for call and P for put
    # S0 - Initial stock price
    # r - Interest rate (constant)
    # tau - Time to maturity
    # K - List of strikes
    # N - Number of expansion terms
    # L - Size of truncation domain (typ.:L=8 or L=10)

    # Reshape K to become a column vector
    if K is not np.array:
        if np.isscalar(K):
            K = np.array([K]).reshape(1, 1)
        else:
            K = np.array(K).reshape(len(K), 1)

    x0 = np.log(S0 / K)

    # Truncation domain
    a = 0.0 - L * np.sqrt(tau)
    b = 0.0 + L * np.sqrt(tau)
    
    # Summation from k = 0 to k=N-1
    k = np.linspace(0,N-1,N).reshape([N,1])
    u = k * np.pi / (b - a);
    # Determine coefficients for put prices
    H_k = CallPutCoefficients(cp,a,b,k)
    mat = np.exp(1j * np.outer((x0 - a) , u))
    temp = cf(u) * H_k
    temp[0] = 0.5 * temp[0]
    value = np.exp(-r * tau) * K * np.real(mat.dot(temp))

    return value


# Determine coefficients for put prices
def CallPutCoefficients(cp,a,b,k):
    if cp == 1:
        c = 0.0
        d = b
        coef = Chi_Psi(a,b,c,d,k)
        Chi_k = coef["chi"]
        Psi_k = coef["psi"]
        if a < b and b < 0.0:
            H_k = np.zeros([len(k),1])
        else:
            H_k = 2.0 / (b - a) * (Chi_k - Psi_k)
    elif cp == -1:
        c = a
        d = 0.0
        coef = Chi_Psi(a,b,c,d,k)
        Chi_k = coef["chi"]
        Psi_k = coef["psi"]
        H_k = 2.0 / (b - a) * (- Chi_k + Psi_k)

    return H_k


def Chi_Psi(a,b,c,d,k):
    psi = np.sin(k * np.pi * (d - a) / (b - a)) - np.sin(k * np.pi * (c - a)/ (b - a))
    psi[1:] = psi[1:] * (b - a) / (k[1:] * np.pi)
    psi[0] = d - c
    
    chi = 1.0 / (1.0 + np.power((k * np.pi / (b - a)) , 2.0))
    expr1 = np.cos(k * np.pi * (d - a)/(b - a)) * np.exp(d) - np.cos(k * np.pi  * (c - a) / (b - a)) * np.exp(c)
    expr2 = k * np.pi / (b - a) * np.sin(k * np.pi * (d - a) / (b - a)) - k * np.pi / (b - a) * np.sin(k * np.pi * (c - a) / (b - a)) * np.exp(c)
    chi = chi * (expr1 + expr2)
    
    value = {"chi":chi,"psi":psi }
    return value

# Black-Scholes call option price
def BS_Call_Option_Price(cp,S_0,K,sigma,tau,r):
    K = np.array(K).reshape([len(K),1])
    d1 = (np.log(S_0 / K) + (r + 0.5 * np.power(sigma,2.0))
    * tau) / float(sigma * np.sqrt(tau))
    d2 = d1 - sigma * np.sqrt(tau)
    if cp == 1:
        value = st.norm.cdf(d1) * S_0 - st.norm.cdf(d2) * K * np.exp(-r * tau)
    elif cp == -1:
        value = st.norm.cdf(-d2) * K * np.exp(-r * tau) - st.norm.cdf(-d1)*S_0
    return value

# Implied volatility method
def ImpliedVolatility(cp,marketPrice,K,T,S_0,r):
    func = lambda sigma: np.power(BS_Call_Option_Price(cp,S_0,K,sigma,T,r) - marketPrice,1.0)
    impliedVol = optimize.newton(func, 0.7, tol=1e-5)
    return impliedVol

def ChFHestonModel(r,tau,kappa,gamma,vbar,v0,rho):
    i = complex(0.0,1.0)
    D1 = lambda u: np.sqrt(np.power(kappa-gamma*rho*i*u,2)+(u*u+i*u)*gamma*gamma)
    g = lambda u: (kappa-gamma*rho*i*u-D1(u))/(kappa-gamma*rho*i*u+D1(u))
    C = lambda u: (1.0-np.exp(-D1(u)*tau))/(gamma*gamma*(1.0-g(u)*D1(u)*tau)) *(kappa-gamma*rho*i*u-D1(u))
    # Note that we exclude the term -r*tau, as the discounting is performed in the COS method
    A = lambda u: r * i*u *tau + kappa*vbar*tau/gamma/gamma *(kappa-gamma*rho*i*u-D1(u)) - 2*kappa*vbar/gamma/gamma*np.log((1-g(u)*np.exp(-D1(u)*tau))/(1-g(u)))
    # Characteristic function for the Heston model
    cf = lambda u: np.exp(A(u) + C(u)*v0)
    return cf



cp = 1
S0 = 100
r = 0.00
tau = 1

K = np.linspace(50,150)
K = np.array(K).reshape([len(K),1])

# COS method settings
L = 3

kappa = 1.5768
gamma = 0.5751
vbar = 0.0398
rho =-0.5711
v0 = 0.0175

# Compute ChF for the Heston model
cf = ChFHestonModel(r,tau,kappa,gamma,vbar,v0,rho)

# The COS method -- Reference value
N = 5000
valCOSRef = CallPutOptionPriceCOSMthd(cf, cp, S0, r, tau, K, N, L)

plt.figure(1)
plt.plot(K,valCOSRef,'-k')
plt.xlabel("strike, K")
plt.ylabel("Option Price")
plt.grid()

# For different numbers of integration terms

N = [32,64,96,128,160,400]
maxError = []
for n in N:
    valCOS = CallPutOptionPriceCOSMthd(cf, cp, S0, r, tau, K, n, L)
    errorMax = np.max(np.abs(valCOS - valCOSRef))
    maxError.append(errorMax)
    print('maximum error for n ={0} is {1}'.format(n,errorMax))
        
    
#%%


cp = 1
S0 = 1.2779
r = 0.0070075
tau = 2

kappa = 1.5
gamma = 0.017132031
vbar = 0.293475456
rho = 0.232453505
v0 = 0.014820628
K = 1.466967

cf = ChFHestonModel(r,tau,kappa,gamma,vbar,v0,rho)
P_COS = CallPutOptionPriceCOSMthd(cf, cp, S0, r, tau, K, n, L)    
print('PCOS',P_COS)



# -*- coding: utf-8 -*-

import numpy as np
import scipy 

def calc_theta(zero_curve,
               mean_reversion_level: float,
               volatility: float,
               n: int=100):
    """
    Per a specified mean_reversion_level (α) and volatility (σ), 
    set theta (θ), based on the term structure of interest rates

    Parameters
    ----------
    alpha : float
        α, the mean reversion parameter of the Hull-White 1 factor model 
    sigma : float
        σ, the (annualised) volatility paramater of the Hull-White 1 factor model 
    n : int, optional
        defines the granularity of the date grid. The default is 100.

    Returns
    -------
    θ_spline_definition : Result of scipy.interpolate.splrep. Input to scipy.interpolate.splev
  
    # To do, to be extended for a term structure of α, σ 
    # Need to specify term structure segmenents if doing that? 

    References:
        [1] Damiano Brigo, Fabio Mercurio - Interest Rate Models Theory and Practice (2001, Springer) 
    """
    
    dt = 1e-5
    α = mean_reversion_level
    σ = volatility
    
    years_grid = np.linspace(zero_curve.zero_data['years'].min(),
                             zero_curve.zero_data['years'].max(),n)
    
    # Calculate the 2nd derivative by numerical differentiation
    f = zero_curve.instantaneous_forward_rate(years=years_grid)
    f_plus_dt = zero_curve.instantaneous_forward_rate(years=years_grid+dt)
    f_minus_dt = zero_curve.instantaneous_forward_rate(years=years_grid-dt)
    df_dt = (f_plus_dt - f_minus_dt) / (2 * dt)

    # Equation (3.34), in Section 3.3.1 'The Short-Rate Dynamics' on page 73 of [1] (page 121 of the pdf)
    θ_grid = df_dt + α * f + (σ**2) * (1-np.exp(-2*α*years_grid)) / (2*α)

    θ_spline_definition = scipy.interpolate.splrep(years_grid, θ_grid) 

    return θ_spline_definition
    

def calc_B(t, T, α):
    # [1] Damiano Brigo, Fabio Mercurio - Interest Rate Models Theory and Practice (2001, Springer) 
    #     In section 3.3.2 'Bond and Option Pricing', page 75 (page 123 of the pdf) in [1]
    # [2] MAFS525 – Computational Methods for Pricing Structured Products, Slide 3/41
    return (1/α) *(1-np.exp(-α*(T- t)))


def calc_A(t, T, θ_spline_definition, α, σ):         
    # MAFS525 – Computational Methods for Pricing Structured Products, Slide 3/41            
    def integrand_1(t, T): 
        return calc_B(t, T)**2    
    def integrand_2(t, T):
        θ_values = scipy.interpolate.splev(t, θ_spline_definition)
        return θ_values * calc_B(t, T)
        
    integrand_1_res = scipy.integrate.quad(integrand_1, t, T)[0]
    integrand_2_res = scipy.integrate.quad(integrand_2, t, T)[0]
    
    return 0.5*(σ**2) * integrand_1_res - integrand_2_res
        

def calc_discount_factor(t, T, θ, α, σ, r0):
    # MAFS525 – Computational Methods for Pricing Structured Products, Slide 3/41 (Hull Whi)
    B = calc_B(t, T, α)
    A = calc_A(t, T, θ, α, σ)
    return np.exp(A - r0*B)



# -*- coding: utf-8 -*

import numpy as np
from scipy.stats import norm


if __name__ == "__main__":
    import os
    import pathlib
    import sys
    
    os.chdir(pathlib.Path(__file__).parent.parent.parent.parent.resolve()) 
    sys.path.append(os.getcwd())
    print('__main__ - current working directory:', os.getcwd())


#%% Test COS method function with the normal distribution

def cos_method(p, cf, dt, a, b, N=160):
    """
    PROBLEM STATEMENT: The charactheristic function (or moment generating function, or
    Laplace transform, from which to retrieve it) of a certain random variable X is known. 
    The objective is approximating f and F, respectively PDF and CDF of X.

    Parameters
    ----------
    p : TYPE
        point at which to evaluate the CDF (F) of f - must be in [a,b].
    cf : TYPE
        characteristic function provided as a function handle where the only variable is the point at which to evaluate it. 
    N : int
        Number of sums, per [1], N=160 the COS method quickly priced options with high accuracy
    dt : float
        Step size for approximating the integral
    a : float
        Lower integration bound
    b : float
        Upper integration bound

    Returns
    -------
    Fp : float
        CDF evaluated at p
    F : float
        CDF evaluated on pts
    f : float
        PDF evaluated on pts
    I : float
        Integral of the PDF over [a,b]: 
    pts : float
        points over which the CDF and PDF are sampled: a:dt:pts (and point p)    
        
        
    References:
    [1] Fang, Fang & Oosterlee, Cornelis. (2008). A Novel Pricing Method for European Options Based on Fourier-Cosine Series Expansions. SIAM J. Scientific Computing. 31. 826-848. 10.1137/080718061. 
    """
    
    assert dt > 0.0    
    assert (a + dt) <= b
    assert p >= a
    assert p <= b
    
    pts = np.sort(np.unique(np.concatenate(([p], np.arange(a, b + dt, dt))))) # Array from [a, a+dt, a+2*dt, ..., b]
    dt_pts = np.append(np.diff(pts), dt)
    k = np.arange(N + 1) # Array from [0, 1, 2, ... N, N+1]

    w = np.append(0.5, np.ones(N))
    
    Fk = 2 / (b-a) * np.real(cf(k * np.pi / (b-a)) * np.exp(-1j * k * np.pi * a / (b-a))) # Equation 8 in [1] 
    C = lambda x: np.cos(k * np.pi * (x - a) / (b-a)) # Inner term of equation 11 in [1] 
    PDF_pts = np.array([np.sum(w * (Fk * C(x))) for x in pts]) # Equation 11 in [1]

    integral_of_a_to_b = np.sum(PDF_pts * dt_pts)
    if integral_of_a_to_b < norm.cdf(5):
        raise ValueError('The integral of the PDF over [a,b] is not close enough to 1. Please increase the range of [a,b]')
    
    CDF_pts = np.cumsum(PDF_pts * dt_pts)
    CDF_p = CDF_pts[np.where(pts == p)[0][0]]

    return CDF_p, CDF_pts, PDF_pts, integral_of_a_to_b, pts

#%%

def heston_cos_vanilla_european(s0, tau, r_f, r_d, cp, K, v0, vv, kappa, theta, rho, N=160, L=10, calculate_via_put_call_parity=True):
    
    """
    Computes the call or put option prices using the COS method.
    
    Parameters:
    - cp (int): 1 for call and -1 for put
    - s0 (float): Initial stock price
    - r (float): Interest rate
    - tau (float): Time to maturity
    - K (list or np.array): List of strike prices
    - N (int): Number of expansion terms (<160 should be sufficient per Fang, 2008)
    - L (float): Size of truncation domain
    
    Returns:
    - np.array: Option prices

    References:
    [1] Fang, Fang & Oosterlee, Cornelis. (2008). A Novel Pricing Method for European Options Based on Fourier-Cosine Series Expansions. SIAM J. Scientific Computing. 31. 826-848. 10.1137/080718061. 
    """     

    x0 = np.log(s0 / K) # Per [1] in section 3.1 on page 6 of 21

    if True:
        # Apply the truncation method per appendix 11 of [1] 
        mu = r_d - r_f # technically we should use the drift implied from market FX forward rate, but the IR parity forward rate should be close enough
        model_param =  {'tau': tau,
                        'mu': mu,
                        'v0': v0,
                        'vv': vv,
                        'kappa': kappa,
                        'theta': theta, 
                        'rho': rho}
        a, b = get_cos_truncation_range(model='heston', L=L, model_param=model_param)
    else:
        # This is a simpler truncation method
        # It produces a wider bound than the more complex method detailed in [1]
        a, b = -L * np.sqrt(tau), L * np.sqrt(tau)

    #print('a,b SF:',a,b)

    # Summation from k = 0 to k = N-1
    k = np.arange(N)
    
    u = (k*np.pi)/(b-a)
    #print("SF omega: ", sum(u))    
    
    chf = chf_heston_fang2008(u=u, tau=tau, r_f=r_f, r_d=r_d, v0=v0, vv=vv, kappa=kappa, theta=theta, rho=rho)
    #print('SF: SumCHF: ',sum(chf))
    
    Uk_call = calculate_Uk_european_options(cp=1, a=a, b=b, k=k) # Per equation 29 of [1] (page 8 of 21)
    Uk_put = calculate_Uk_european_options(cp=-1, a=a, b=b, k=k) # Per equation 29 of [1] (page 8 of 21)
    #print('Uk (c,p) SF:',sum(Uk_call), sum(Uk_put))
    
    Fk = np.real(chf  * np.exp(1j * k * np.pi * (x0 - a)/(b-a))) 
    Fk[0] = 0.5 * Fk[0] # Per page 3/21 of [1], "where Σ′ indicates that the first term in the summation is weighted by one-half"    
    
    # Note for the equations below, K can adjusted to be a vector, per equation 34 of [1] (page 8 of 21)
    if cp == -1 or calculate_via_put_call_parity:
        put_px = K * np.sum(np.multiply(Fk, Uk_put)) * np.exp(-r_d * tau) 
    if cp == 1 and calculate_via_put_call_parity:
        call_px = put_px + s0 * np.exp(-r_f * tau) - K * np.exp(-r_d * tau) # Method by put-call parity is more stable
    elif cp == 1 and not calculate_via_put_call_parity:
        call_px = K * np.sum(np.multiply(Fk, Uk_call)) * np.exp(-r_d * tau) 
    
    if cp == 1:
        return call_px
    elif cp == -1:
        return put_px
    


#%% Misc 


def get_cos_truncation_range(model, L, model_param):
    """
    # This function returns the truncation range of the characteristic function 
    # Defined for COS method approximations of stochastic models as detailed in Appendix of Fang & Oosterlee (2008), in Table 11, page 21/21.
    # Note: c1, c2, ... cn are the 1st, 2nd, .. nth cumulant of ln(ST/K)
    # References:
    # [1] Fang, Fang & Oosterlee, Cornelis. (2008). A Novel Pricing Method for European Options Based on Fourier-Cosine Series Expansions. SIAM J. Scientific Computing. 31. 826-848. 10.1137/080718061. 
        
    Parameters
    ----------
    model : str
        the type of stochastic model (geomentric_brownion_motion, heston, VG, CGMY).
    model_param : dict
        defining paramaters of the stochastic model

    Returns
    -------
    a : float
        lower truncation value.
    b : float
        upper truncation value.

    """
    
    # From Table 11 in the Appendix A of [1]
    if model == 'geomentric_brownion_motion':
        # GBM is defined by two parameters:
        # mu: the drift
        # σ: the volatility
        mu,σ,tau = model_param
        
        c1 = mu * tau
        c2 = (σ**2) * tau
        c4 = 0
     
    elif model == 'heston':
        # Remap the parameters per the symbols used in [1] so comparison to the formulae on page 21 of 21 is easier
        tau = model_param['tau'] # time-to-expiry
        mu = model_param['mu'] # drift mu. mu = r-q. For FXO mu = r_DOM - r_FOR
        u0 = model_param['v0'] # Initial volatility.
        η  = model_param['vv'] # Volatility of volatility.
        λ = model_param['kappa'] # rate of mean reversion to the long-run variance
        u_bar = model_param['theta'] # Long-run variance
        rho = model_param['rho'] # Correlation.
                        
        c1 = mu * tau + (1- np.exp(-1 * λ * tau)) * (u_bar - u0) / (2 * λ) - 0.5 * u_bar * tau
        
        c2 = (1 / (8 * λ**3)) * (
            (η * tau * λ * np.exp(-λ * tau) * (u0 - u_bar) * (8 * λ * rho - 4 * η))
            + λ * rho * η * (1 - np.exp(-λ * tau)) * (16 * u_bar - 8 * u0)
            + 2 * u_bar * λ * tau * (-4 * λ * rho * η + η**2 + 4 * λ**2)
            + η**2 * ((u_bar - 2 * u0) * np.exp(-2 * λ * tau) + u_bar * (6 * np.exp(-λ * tau) - 7) + 2 * u0)
            + 8 * λ**2 * (u0 - u_bar) * (1 - np.exp(-λ * tau))
        )
                
        c4 = 0
        
        #print("SF: c1,c2", c1, c2)
          
    elif model == 'VG':
        # Detailed in paper but not priority for this project
        pass
    elif model == 'CGMY':
        # Detailed in paper but not priority for this project
        pass    
        
    a = c1 - L * np.sqrt(np.abs(c2) + np.sqrt(np.abs(c4))) # Per Equantion 49 from [1]
    b = c1 + L * np.sqrt(np.abs(c2) + np.sqrt(np.abs(c4))) # Per Equantion 49 from [1]

    return a,b


def calculate_Uk_european_options(cp, a, b, k):
    """
    Compute Uk for valuing vanilla european call or put options under Heston using COS method.
    
    Parameters:
    - cp (int): 1 for call, -1 for put
    - a (float): Lower truncation boundary
    - b (float): Upper truncation boundary
    - k (np.array): Array of k values, [0, 1, 2, .... to N] where N is the number of expansion terms
    
    Returns:
    - np.array: Coefficients
    
    References:
    [1] Fang, Fang & Oosterlee, Cornelis. (2008). A Novel Pricing Method for European Options Based on Fourier-Cosine Series Expansions. SIAM J. Scientific Computing. 31. 826-848. 10.1137/080718061. 
    """     

    def calc_chi(k, c, d):
        # Calculate χ, chi, Per equation 22 from [1] (page 6 of 21)
        χ = np.zeros(shape=k.shape)
        term1 = 1.0 / (1.0 + np.power((k * np.pi / (b - a)) , 2.0))
        term2 = np.cos(k * np.pi * (d - a)/(b - a)) * np.exp(d)
        term3 = np.cos(k * np.pi  * (c - a) / (b - a)) * np.exp(c)
        term4 = k * np.pi / (b - a) * np.sin(k * np.pi * (d - a) / (b - a))
        term5 = k * np.pi / (b - a) * np.sin(k * np.pi * (c - a) / (b - a)) * np.exp(c)
        χ = term1 * (term2 - term3 + term4 - term5)
        return χ
        
    def calc_psi(k, c, d):
        # Calculate Ψ, psi, per equation 23 from [1] (page 6 of 21)
        Ψ = np.zeros(shape=k.shape)        
        Ψ[1:] = (np.sin(k[1:] * np.pi * (d-a) / (b-a)) - np.sin(k[1:] * np.pi * (c-a) / (b-a))) * (b-a) / (k[1:] * np.pi)
        Ψ[0] = d - c
        return Ψ
    
    if cp == 1:
        # For call, c=0, d=b per equation 29 from [1] (page 7 of 21)
        χ = calc_chi(k=k, c=0, d=b)
        Ψ = calc_psi(k=k, c=0, d=b)      
        Uk_call = (2 / (b-a)) * (χ - Ψ)
        return Uk_call
    elif cp == -1:
        # For put, c=a, d]0, per equation 29 from [1] (page 7 of 21)
        χ = calc_chi(k=k, c=a, d=0)
        Ψ = calc_psi(k=k, c=a, d=0)
        Uk_put = (2 / (b-a)) * (-χ + Ψ)
        return Uk_put









# -*- coding: utf-8 -*-


if __name__ == "__main__":
    import os
    import pathlib
    os.chdir(pathlib.Path(__file__).parent.parent.parent.parent.resolve())     
    print('__main__ - current working directory:', os.getcwd())
    
import numpy as np
from numba import jit, prange
from scipy.stats import norm
import timeit
import time

from frm.frm.pricing_engine.monte_carlo_generic import generate_rand_nbs


def normalcorr(C: np.array, 
               rand_nbs: np.array):
    """
    Generate correlated pseudo random normal variates using the Cholesky factorization.
    
    Parameters:
    C (np.ndarray): Correlation matrix.
    rand_nbs (np.ndarray): Matrix of normally distributed pseudorandom numbers.
    
    Returns:
    np.ndarray: Correlated pseudo random normal variates.
    """    
    
    M = np.linalg.cholesky(C).T
    return (M @ rand_nbs.T).T


def simulate_heston_single(S0: float, 
                           mu: float, 
                           var0: float, 
                           vv: float, 
                           kappa: float, 
                           theta: float, 
                           rho: float, 
                           tau: float, 
                           rand_nbs: np.array,
                           method: str='quadratic_exponential'):
    
    """
      Simulate trajectories of the spot price and volatility processes in the Heston model.
     
      Parameters:
      S0 (float): initial spot price.
      mu (float): drift (r-q)
      var0 (float): initial variance.
      vv (float): volatility of volatility.
      kappa (float): speed of mean reversion for volatility.
      theta (float): long-run variance
      rho (float): correlation between spot price and volatility.
      tau (float): time end point
      dt (float): the time step size
      rand_nbs (np.ndarray, optional): 2-column array of normally distributed pseudorandom numbers. Defaults to None.
      simulation_method (str, optional): Defaults to 'quadratic_exponential'.
          {'quadratic_exponential,'euler_with_absorption_of_volatility_process','euler_with_reflection_of_volatility_process'}
      Returns:
      np.ndarray: 2-column array containing the simulated trajectories of the spot price and volatility.
      
      References:
      [1] Janek, A., Kluge, T., Weron, R., Wystup, U. (2010). "FX smile in the Heston model"
      """    
     
    dt = tau / nb_timesteps

    t = np.arange(0, nb_timesteps + 1)
    x = np.zeros((nb_timesteps + 1, 2))
    
    C = np.array([[1, rho], [rho, 1]])
    u = normalcorr(C, rand_nbs) * np.sqrt(dt)
    
    if method == 'quadratic_exponential':
        x[0, :] = [np.log(S0), var0]
        phiC = 1.5
        
        for i in range(1, nb_timesteps + 1):
            m = theta + (x[i - 1, 1] - theta) * np.exp(-kappa * dt)
            s2 = (x[i - 1, 1] * vv ** 2 * np.exp(-kappa * dt) / kappa * (1 - np.exp(-kappa * dt)) +
                  theta * vv ** 2 / (2 * kappa) * (1 - np.exp(-kappa * dt)) ** 2)
            phi = s2 / m ** 2
            gamma1 = gamma2 = 0.5
            K0 = -rho * kappa * theta / vv * dt
            K1 = gamma1 * dt * (kappa * rho / vv - 0.5) - rho / vv
            K2 = gamma2 * dt * (kappa * rho / vv - 0.5) + rho / vv
            K3 = gamma1 * dt * (1 - rho ** 2)
            K4 = gamma2 * dt * (1 - rho ** 2)
            
            if phi <= phiC:
                b2 = 2 / phi - 1 + np.sqrt(2 / phi * (2 / phi - 1))
                a = m / (1 + b2)
                x[i, 1] = a * (np.sqrt(b2) + rand_nbs[i - 1, 1]) ** 2
            else:
                p = (phi - 1) / (phi + 1)
                beta = (1 - p) / m
                if 0 <= norm.cdf(rand_nbs[i - 1, 1]) <= p:
                    x[i, 1] = 0
                elif p < norm.cdf(rand_nbs[i - 1, 1]) <= 1:
                    x[i, 1] = 1 / beta * np.log((1 - p) / (1 - norm.cdf(rand_nbs[i - 1, 1])))
            
            x[i, 0] = x[i - 1, 0] + mu * dt + K0 + K1 * x[i - 1, 1] + K2 * x[i, 1] + \
                      np.sqrt(K3 * x[i - 1, 1] + K4 * x[i, 1]) * rand_nbs[i - 1, 0]
        
        x[:, 0] = np.exp(x[:, 0])
    elif method[:5] == 'euler':
        x[0, :] = [S0, v0]
        for i in range(1, nb_timesteps+1):
            if method == 'euler_with_absorption_of_volatility_process':
                x[i - 1, 1] = max(x[i - 1, 1], 0) 
            elif method == 'euler_with_reflection_of_volatility_process':
                x[i - 1, 1] = abs(x[i - 1, 1])
                
            x[i, 1] = x[i - 1, 1] + kappa * (theta - x[i - 1, 1]) * dt + vv * np.sqrt(x[i - 1, 1]) * u[i - 1, 1]
            x[i, 0] = x[i - 1, 0] + x[i - 1, 0] * (mu * dt + np.sqrt(x[i - 1, 1]) * u[i - 1, 0])
    
    return x


def simulate_heston(S0: float, 
                    mu: float, 
                    v0: float, 
                    vv: float, 
                    kappa: float, 
                    theta: float, 
                    rho: float, 
                    tau: float, 
                    rand_nbs: np.array,
                    method: str='quadratic_exponential'):
    """
     Simulate trajectories of the spot price and volatility processes in the Heston model.
     
     Parameters:
     S0 (float): initial spot price.
     mu (float): drift.
     v0 (float): initial volatility.
     vv (float): volatility of volatility.
     kappa (float): speed of mean reversion for volatility.
     theta (float): long-term mean of volatility.
     rho (float): correlation between spot price and volatility.
     tau (float): time end point
     nb_timesteps (float): the number of timesteps in the simulation
     rand_nbs (np.ndarray, optional): 2-column array of normally distributed pseudorandom numbers. Defaults to None.
     simulation_method (str, optional): Defaults to 'quadratic_exponential'.
         {'quadratic_exponential,'euler_with_absorption_of_volatility_process','euler_with_reflection_of_volatility_process'}
     Returns:
     np.ndarray: 2-column array containing the simulated trajectories of the spot price and volatility.
     """    
     
    np.random.seed(0)
    
    nb_timesteps = rand_nbs.shape[0]
    dt = tau / nb_timesteps

    t = np.arange(0, nb_timesteps + 1)
    x = np.zeros((nb_timesteps+1, 2, rand_nbs.shape[2]))
    
    C = np.array([[1, rho], [rho, 1]])
    u = normalcorr(C, rand_nbs) * np.sqrt(dt) # need to check this correlation works as expected
    
    if method == 'quadratic_exponential':
        x[0, 0, :] = np.log(S0)
        x[0, 1, :] = v0
        
        phiC = 1.5
        
        for i in range(1, nb_timesteps + 1):
            m = theta + (x[i - 1, 1, :] - theta) * np.exp(-kappa * dt)
            s2 = (x[i - 1, 1, :] * vv ** 2 * np.exp(-kappa * dt) / kappa * (1 - np.exp(-kappa * dt)) +
                  theta * vv ** 2 / (2 * kappa) * (1 - np.exp(-kappa * dt)) ** 2)
            phi = s2 / m ** 2
            gamma1 = gamma2 = 0.5
            K0 = -rho * kappa * theta / vv * dt
            K1 = gamma1 * dt * (kappa * rho / vv - 0.5) - rho / vv
            K2 = gamma2 * dt * (kappa * rho / vv - 0.5) + rho / vv
            K3 = gamma1 * dt * (1 - rho ** 2)
            K4 = gamma2 * dt * (1 - rho ** 2)
            
            mask = phi <= phiC
            b2 = np.full(mask.shape, np.nan)
            b2[mask] = 2 / phi[mask] - 1 + np.sqrt(2 / phi[mask] * (2 / phi[mask] - 1))
            a = np.full(mask.shape, np.nan)
            a[mask] = m[mask] / (1 + b2[mask])
            x[i, 1, mask] = a[mask] * (np.sqrt(b2[mask]) + rand_nbs[i - 1, 1, mask]) ** 2
            
            mask = np.logical_not(mask)
            p = np.full(mask.shape, np.nan)
            p[mask] = (phi[mask] - 1) / (phi[mask] + 1)
            beta = np.full(mask.shape, np.nan)
            beta[mask] = (1 - p[mask]) / m[mask]
            
            mask_0_p = np.logical_and(0 <= norm.cdf(rand_nbs[i - 1, 1, :]), norm.cdf(rand_nbs[i - 1, 1, :]) <= p)
            mask_net = np.logical_and(mask, mask_0_p)
            x[i, 1, mask_net] = 0
            
            mask_p_1 = np.logical_and(p < norm.cdf(rand_nbs[i - 1, 1, :]), norm.cdf(rand_nbs[i - 1, 1, :]) <= 1)
            mask_net = np.logical_and(mask, mask_p_1)
            x[i, 1, mask_net] = 1 / beta[mask_net] * np.log((1 - p[mask_net]) / (1 - norm.cdf(rand_nbs[i - 1, 1, mask_net])))
                        
            x[i, 0, :] = x[i - 1, 0, :] + mu * dt + K0 + K1 * x[i - 1, 1, :] + K2 * x[i, 1, :] + \
                      np.sqrt(K3 * x[i - 1, 1, :] + K4 * x[i, 1, :]) * rand_nbs[i - 1, 0, :]
        
        x[:, 0] = np.exp(x[:, 0])
    elif method[:5] == 'euler':
        x[0, 0, :] = S0
        x[0, 1, :] = v0
        
        for i in range(1, nb_timesteps + 1):
            if method == 'euler_with_absorption_of_volatility_process':
                x[i - 1, 1, :] = np.maximum(0, x[i - 1, 1, :])
            elif method == 'euler_with_reflection_of_volatility_process':
                x[i - 1, 1, :] = np.abs(x[i - 1, 1, :])

            x[i, 1, :] = x[i - 1, 1, :] + kappa * (theta - x[i - 1, 1, :]) * dt + vv * np.sqrt(x[i - 1, 1, :]) * u[i - 1, 1, :]
            x[i, 0, :] = x[i - 1, 0, :] + x[i - 1, 0, :] * (mu * dt + np.sqrt(x[i - 1, 1, :]) * u[i - 1, 0, :])
    
    return x


#%%

if __name__ == '__main__':
    
    # This is a check to demonstate the vectorised function gives the same result as the non vectorised function
    
    np.random.seed(0)
    
    S0 = 0.6629
    mu = 0
    v0 = 0.01030476434426229
    vv = 0.2992984338043174
    kappa = 1.5
    theta = 0.013836406947876231
    rho = -0.3432643651463818
    tau = 2
    nb_timesteps = 100
    
    rand_nbs = generate_rand_nbs(nb_timesteps=nb_timesteps,
                                 nb_rand_vars=2,
                                 nb_simulations=1 * 1000,
                                 flag_apply_antithetic_variates=False)
   
    
    t1 = time.time()
    
    result = simulate_heston(S0=S0,
        mu=mu,
        v0=v0,
        vv=vv,
        kappa=kappa,
        theta=theta,
        rho=rho, 
        tau=tau, 
        rand_nbs=rand_nbs,
        method='quadratic_exponential')
    
    t2 = time.time()

    if True:
        for i in range(rand_nbs.shape[2]):
            
            print(i, '/', rand_nbs.shape[2])
    
            rand_nbs_single = rand_nbs[:,:,i]
    
            results_single = simulate_heston_single(S0=S0,
                mu=mu,
                v0=v0,
                vv=vv,
                kappa=kappa,
                theta=theta,
                rho=rho, 
                tau=tau, 
                rand_nbs=rand_nbs_single,
                method='quadratic_exponential')
                
            check = (result[:,:,i] - results_single).sum()
            epsilon = 1e-8
            assert check < epsilon
            
        t3 = time.time()
        
        print('Runtime of vectorised function:', t2-t1)
        print('Runtime of for-loop function:', t3-t2)
    
    result_avg = result.mean(axis=2)









# -*- coding: utf-8 -*-


if __name__ == "__main__":
    import os
    import pathlib
    os.chdir(pathlib.Path(__file__).parent.parent.parent.resolve())     
    print('__main__ - current working directory:', os.getcwd())

import re 
import unicodedata
import logging
import numpy as np
import pandas as pd
from pandas import DateOffset
from typing import Union
import datetime as dt
  

def get_spot_offset(curve_ccy: str=None) -> int:
    
    if curve_ccy == None:
        return 2
    
    if not isinstance(curve_ccy, str):
        logging.error("function input 'curve_ccy' must be a string")
        raise TypeError("function input 'curve_ccy' must be a string")    
    
    if len(curve_ccy) == 6:
        # http://www.londonfx.co.uk/valdates.html
        if curve_ccy in {'usdcad','cadusd',
                         'usdphp','phpusd',
                         'usdrub','rubusd',
                         'usdtry','tryusd'}:
            return 1
        else:
            return 2

    # Spot offset for interest rate products
    elif len(curve_ccy) == 3:    
        if curve_ccy in {'aud','cad'}:
            return 1
        elif curve_ccy in {'nzd','jpy','usd','eur'}:
            return 2
        elif curve_ccy in {'gbp'}:
            return 0
        else:
            return 2
        
    else:
        return 2
        #logging.error("invalid 'curve_ccy' value: " + curve_ccy)      
        #raise ValueError("invalid 'curve_ccy' value: " + curve_ccy)


@np.vectorize
def clean_tenor(tenor: str) -> str:
    if not isinstance(tenor, str):
        logging.error("function input 'tenor' must be a string")
        raise TypeError("function input 'tenor' must be a string")      
    
    tenor = unicodedata.normalize('NFKD', tenor)

    tenor = tenor.lower().replace(' ','').replace('/','').replace('\n', '').replace('\r', '')
    
    tenor = tenor.replace('days','d')
    tenor = tenor.replace('day','d')
    tenor = tenor.replace('weeks','w')
    tenor = tenor.replace('week','w')
    tenor = tenor.replace('months','m')
    tenor = tenor.replace('month','m')
    tenor = tenor.replace('mon','m')
    tenor = tenor.replace('years','y')    
    tenor = tenor.replace('year','y')
    tenor = tenor.replace('yrs','y')        
    tenor = tenor.replace('yr','y')     
    
    tenor = tenor.replace('overnight','on') 
    tenor = tenor.replace('tomorrownext','tn')
    tenor = tenor.replace('tomnext','sn')  
    tenor = tenor.replace('spotweek','sw')    
    tenor = tenor.replace('spotnext','sn')     
    tenor = tenor.replace('spot','sp') 
    return tenor 
   

@np.vectorize
def tenor_name_to_date_offset(tenor_name: str) -> pd.DateOffset:    
    if not isinstance(tenor_name, str):
        logging.error("function input 'tenor' must be a string")
        raise TypeError("function input 'tenor' must be a string")      
        
    misc_tenors_offset = {
        'on' : DateOffset(days=1), # 1 day (overnight)
        'tn' : DateOffset(days=2), # 2 days (tomorrow next)    
        'sp'  : DateOffset(days=0), # spot date
        'sn' : DateOffset(days=1), # 1 day (spot next)
        'sw' : DateOffset(days=5), # 5 days (spot week)        
        }
    
    if tenor_name in misc_tenors_offset.keys():
        offset = misc_tenors_offset[tenor_name]
    else:
        # Identity tenors specified in integer days only; 1D, 30D , 360D
        if re.search('^\d+d$',tenor_name) is not None:
            offset = DateOffset(days=int(tenor_name[:-1]))
        # Identity tenors specified in integer weeks only; 1W, 52W, 104W
        elif re.search('^\d+w$',tenor_name) is not None:
            offset = DateOffset(weeks=int(tenor_name[:-1]))
        # Identity tenors specified in integer months only; 1M, 12M, 120M
        elif re.search('^\d+m$',tenor_name) is not None:
            offset = DateOffset(months=int(tenor_name[:-1]))       
        # Identity tenors specified in integer years only; 1Y, 10Y, 100Y
        elif re.search('^\d+y$',tenor_name) is not None:
            offset = DateOffset(years=int(tenor_name[:-1]))   
        # Identity tenors specified in integer years and integer monthly only; 1Y3M, 10Y6M, 100Y1M
        elif re.search('^\d+y\d+m$',tenor_name) is not None:
            years,months = tenor_name[:-1].split('y')
            total_months = int(years)*12 + int(months)
            offset = DateOffset(months=total_months)   
        else: 
            logging.error("invalid 'tenor' value: " + tenor_name)      
            raise ValueError("invalid 'tenor' value: " + tenor_name)
        
    return offset

@np.vectorize
def offset_market_data_date(curve_date: np.datetime64,
                spot_date: np.datetime64,
                tenor_name: str) -> np.datetime64:    
    
    # offset from 
    # (i) the curve_date for the ON and TN tenors and from;
    # (ii) the spot date for all other tenors   
    
    date_offset = tenor_name_to_date_offset(tenor_name)
    
    if tenor_name in {'on'}:
        return curve_date + date_offset.item() 
    else:
        return spot_date + date_offset.item()

@np.vectorize
def get_tenor_effective_date(tenor_name, curve_date, spot_date):
    if tenor_name in {'on'}:
        return curve_date 
    else:
        return spot_date 

def calc_tenor_date(curve_date: pd.Timestamp,
                    tenor_name: Union[str, np.ndarray], 
                    curve_ccy: str=None, 
                    holiday_calendar: np.busdaycalendar=None,
                    spot_offset: bool=True):
        
    if holiday_calendar == None or pd.isna(holiday_calendar):
        holiday_calendar = np.busdaycalendar()
    
    if type(curve_date) == dt.date:
        curve_date = np.datetime64(curve_date)
    else:
        curve_date = np.datetime64(curve_date.date())
            
    if spot_offset:
        spot_date = np.busday_offset(curve_date, offsets=get_spot_offset(curve_ccy), roll='following', busdaycal=holiday_calendar)
    else: 
        spot_date = curve_date
        
    cleaned_tenor_name = clean_tenor(tenor_name)
    if cleaned_tenor_name.size == 1:
        cleaned_tenor_name = cleaned_tenor_name.item()
    
    offset_date = offset_market_data_date(curve_date, spot_date, cleaned_tenor_name)
    
    holiday_rolled_offset_date = np.busday_offset(offset_date.astype('datetime64[D]'), offsets=0, roll='following', busdaycal=holiday_calendar)
    
    if holiday_rolled_offset_date.shape == ():
        holiday_rolled_offset_date = pd.Timestamp(holiday_rolled_offset_date.item()) # For scalar
    else:
        holiday_rolled_offset_date = pd.DatetimeIndex(holiday_rolled_offset_date) # For array
    
    
    
    return holiday_rolled_offset_date, cleaned_tenor_name, pd.Timestamp(spot_date)  
        



# -*- coding: utf-8 -*-


USD_LIBOR_SWAP_1Y = {
    # General key terms
    'notional_currency': 'USD',
    'city_holidays': 'new_york',
     # Pay fixed leg key terms
    'pay_payment_frequency':'A',
    'pay_day_count_basis':'ACT/360',
     # Receive float leg key terms
    'rec_payment_frequency':'Q',
    'rec_day_count_basis':'ACT/360',
    'rec_float_spread':0.0}

USD_LIBOR_SWAP_ABOVE_1Y = {
    # General key terms
    'notional_currency': 'USD',
    'city_holidays': 'new_york',
     # Pay fixed leg key terms
    'pay_payment_frequency':'S',
    'pay_day_count_basis':'30/360',
     # Receive float leg key terms
    'rec_payment_frequency':'Q',
    'rec_day_count_basis':'ACT/360',
    'rec_float_spread':0.0}

USD_SOFR = {
    # General key terms
    'notional_currency': 'USD',
    'city_holidays': 'new_york',
     # Pay fixed leg key terms
    'pay_payment_frequency':'A',
    'pay_day_count_basis':'ACT/360',
     # Receive float leg key terms
    'rec_payment_frequency':'A',
    'rec_day_count_basis':'ACT/365',
    'rec_float_spread':0.0}

JPY_LIBOR_SWAP = {
    # General key terms
    'notional_currency': 'JPY',
    'city_holidays': 'tokyo',
     # Pay fixed leg key terms
    'pay_payment_frequency':'S',
    'pay_day_count_basis':'ACT/360',
     # Receive float leg key terms
    'rec_payment_frequency':'S',
    'rec_day_count_basis':'ACT/36S',
    'rec_float_spread':0.0}

NZD_BKBM_SWAP = {
    # General key terms
    'notional_currency': 'NZD',
    'city_holidays': 'wellington',
     # Pay fixed leg key terms
    'pay_payment_frequency':'S',
    'pay_day_count_basis':'ACT/365',
     # Receive float leg key terms
    'rec_payment_frequency':'Q',
    'rec_day_count_basis':'ACT/365',
    'rec_float_spread':0.0}

AUD_BBSW_SWAP_3Y_AND_UNDER = {
    # General key terms
    'notional_currency': 'AUD',
    'city_holidays': 'sydney',
     # Pay fixed leg key terms
    'pay_payment_frequency':'Q',
    'pay_day_count_basis':'ACT/365',
     # Receive float leg key terms
    'rec_payment_frequency':'Q',
    'rec_day_count_basis':'ACT/365',
    'rec_float_spread':0.0}

AUD_BBSW_SWAP_ABOVE_3Y = {
    # General key terms
    'notional_currency': 'AUD',
    'city_holidays': 'sydney',
     # Pay fixed leg key terms
    'pay_payment_frequency':'S',
    'pay_day_count_basis':'ACT/365',
     # Receive float leg key terms
    'rec_payment_frequency':'',
    'rec_day_count_basis':'ACT/365',
    'rec_float_spread':0.0}